# Copyright (c) 2018-2019 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
AuxIVA
======

Blind Source Separation using independent vector analysis based on auxiliary function.
This function will separate the input signal into statistically independent sources
without using any prior information.

The algorithm in the determined case, i.e., when the number of sources is equal to
the number of microphones, is AuxIVA [1]_. When there are more microphones (the overdetermined case),
a computationaly cheaper variant (OverIVA) is used [2]_.

Example
-------
.. code-block:: python

    from scipy.io import wavfile
    import pyroomacoustics as pra

    # read multichannel wav file
    # audio.shape == (nsamples, nchannels)
    fs, audio = wavfile.read("my_multichannel_audio.wav")

    # STFT analysis parameters
    fft_size = 4096  # `fft_size / fs` should be ~RT60
    hop == fft_size // 2  # half-overlap
    win_a = pra.hann(fft_size)  # analysis window
    # optimal synthesis window
    win_s = pra.transform.compute_synthesis_window(win_a, hop)

    # STFT
    # X.shape == (nframes, nfrequencies, nchannels)
    X = pra.transform.analysis(audio, fft_size, hop, win=win_a)

    # Separation
    Y = pra.bss.auxiva(X, n_iter=20)

    # iSTFT (introduces an offset of `hop` samples)
    # y contains the time domain separated signals
    # y.shape == (new_nsamples, nchannels)
    y = pra.transform.synthesis(Y, fft_size, hop, win=win_s)

References
----------

.. [1] N. Ono, *Stable and fast update rules for independent vector analysis based
    on auxiliary function technique,* Proc. IEEE, WASPAA, pp. 189-192, Oct. 2011.

.. [2] R. Scheibler and N. Ono, Independent Vector Analysis with more Microphones
    than Sources, arXiv, 2019.  https://arxiv.org/abs/1905.07880
"""
import numpy as np

from .common import projection_back


def auxiva(
    X,
    n_src=None,
    n_iter=20,
    proj_back=True,
    W0=None,
    model="laplace",
    init_eig=False,
    return_filters=False,
    callback=None,
):

    """
    This is an implementation of AuxIVA/OverIVA that separates the input
    signal into statistically independent sources. The separation is done
    in the time-frequency domain and the FFT length should be approximately
    equal to the reverberation time.

    Two different statistical models (Laplace or time-varying Gauss) can
    be used by using the keyword argument `model`. The performance of Gauss
    model is higher in good conditions (few sources, low noise), but Laplace
    (the default) is more robust in general.

    Parameters
    ----------
    X: ndarray (nframes, nfrequencies, nchannels)
        STFT representation of the signal
    n_src: int, optional
        The number of sources or independent components. When
        ``n_src==nchannels``, the algorithms is identical to AuxIVA. When
        ``n_src==1``, then it is doing independent vector extraction.
    n_iter: int, optional
        The number of iterations (default 20)
    proj_back: bool, optional
        Scaling on first mic by back projection (default True)
    W0: ndarray (nfrequencies, nsrc, nchannels), optional
        Initial value for demixing matrix
    model: str
        The model of source distribution 'gauss' or 'laplace' (default)
    init_eig: bool, optional (default ``False``)
        If ``True``, and if ``W0 is None``, then the weights are initialized
        using the principal eigenvectors of the covariance matrix of the input
        data. When ``False``, the demixing matrices are initialized with identity
        matrix.
    return_filters: bool
        If true, the function will return the demixing matrix too
    callback: func
        A callback function called every 10 iterations, allows to monitor
        convergence

    Returns
    -------
    Returns an (nframes, nfrequencies, nsources) array. Also returns
    the demixing matrix (nfrequencies, nchannels, nsources)
    if ``return_values`` keyword is True.
    """

    n_frames, n_freq, n_chan = X.shape

    # default to determined case
    if n_src is None:
        n_src = n_chan

    assert (
        n_src <= n_chan
    ), "The number of sources cannot be more than the number of channels."

    if model not in ["laplace", "gauss"]:
        raise ValueError("Model should be either ""laplace"" or ""gauss"".")

    # covariance matrix of input signal (n_freq, n_chan, n_chan)
    Cx = np.mean(X[:, :, :, None] * np.conj(X[:, :, None, :]), axis=0)

    W_hat = np.zeros((n_freq, n_chan, n_chan), dtype=X.dtype)
    W = W_hat[:, :n_src, :]
    J = W_hat[:, n_src:, :n_src]

    def tensor_H(T):
        return np.conj(T).swapaxes(1, 2)

    def update_J_from_orth_const():
        tmp = np.matmul(W, Cx)
        J[:, :, :] = tensor_H(np.linalg.solve(tmp[:, :, :n_src], tmp[:, :, n_src:]))

    # initialize A and W
    if W0 is None:

        if init_eig:
            # Initialize the demixing matrices with the principal
            # eigenvectors of the input covariance
            v, w = np.linalg.eig(Cx)
            for f in range(n_freq):
                ind = np.argsort(v[f])[-n_src:]
                W[f, :, :] = np.conj(w[f][:, ind])

        else:
            # Or with identity
            for f in range(n_freq):
                W[f, :, :n_src] = np.eye(n_src)

    else:
        W[:, :, :] = W0

    # We still need to initialize the rest of the matrix
    if n_src < n_chan:
        update_J_from_orth_const()
        for f in range(n_freq):
            W_hat[f, n_src:, n_src:] = -np.eye(n_chan - n_src)

    eps = 1e-15
    eyes = np.tile(np.eye(n_chan, n_chan), (n_freq, 1, 1))
    V = np.zeros((n_freq, n_chan, n_chan), dtype=X.dtype)
    r_inv = np.zeros((n_src, n_frames))
    r = np.zeros((n_src, n_frames))

    # Things are more efficient when the frequencies are over the first axis
    Y = np.zeros((n_freq, n_src, n_frames), dtype=X.dtype)
    X_original = X
    X = X.transpose([1, 2, 0]).copy()

    # Compute the demixed output
    def demix(Y, X, W):
        Y[:, :, :] = np.matmul(W, X)

    for epoch in range(n_iter):

        demix(Y, X, W)

        if callback is not None and epoch % 10 == 0:
            Y_tmp = Y.transpose([2, 0, 1])
            if proj_back:
                z = projection_back(Y_tmp, X_original[:, :, 0])
                callback(Y_tmp * np.conj(z[None, :, :]))
            else:
                callback(Y_tmp)

        # shape: (n_frames, n_src)
        if model == "laplace":
            r[:, :] = 2.0 * np.linalg.norm(Y, axis=0)
        elif model == "gauss":
            r[:, :] = (np.linalg.norm(Y, axis=0) ** 2) / n_freq

        # ensure some numerical stability
        r[r < eps] = eps

        r_inv[:, :] = 1.0 / r

        # Update now the demixing matrix
        for s in range(n_src):
            # Compute Auxiliary Variable
            # shape: (n_freq, n_chan, n_chan)
            V[:, :, :] = np.matmul(
                (X * r_inv[None, s, None, :]), np.conj(X.swapaxes(1, 2)) / n_frames
            )

            WV = np.matmul(W_hat, V)
            W[:, s, :] = np.conj(np.linalg.solve(WV, eyes[:, :, s]))

            # normalize
            denom = np.matmul(
                np.matmul(W[:, None, s, :], V[:, :, :]), np.conj(W[:, s, :, None])
            )
            W[:, s, :] /= np.sqrt(denom[:, :, 0])

            # Update the mixing matrix according to orthogonal constraints
            if n_src < n_chan:
                update_J_from_orth_const()

    demix(Y, X, W)

    Y = Y.transpose([2, 0, 1]).copy()

    if proj_back:
        z = projection_back(Y, X_original[:, :, 0])
        Y *= np.conj(z[None, :, :])

    if return_filters:
        return Y, W
    else:
        return Y



# Copyright (c) 2018-2019 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
RT60 Measurement Routine
========================

Automatically determines the reverberation time of an impulse response
using the Schroeder method [1]_.

References
----------

.. [1] M. R. Schroeder, "New Method of Measuring Reverberation Time,"
    J. Acoust. Soc. Am., vol. 37, no. 3, pp. 409-412, Mar. 1968.
"""
import numpy as np


def measure_rt60(h, fs=1, decay_db=60, plot=False, rt60_tgt=None):
    """
    Analyze the RT60 of an impulse response. Optionaly plots some useful information.

    Parameters
    ----------
    h: array_like
        The impulse response.
    fs: float or int, optional
        The sampling frequency of h (default to 1, i.e., samples).
    decay_db: float or int, optional
        The decay in decibels for which we actually estimate the time. Although
        we would like to estimate the RT60, it might not be practical. Instead,
        we measure the RT20 or RT30 and extrapolate to RT60.
    plot: bool, optional
        If set to ``True``, the power decay and different estimated values will
        be plotted (default False).
    rt60_tgt: float
        This parameter can be used to indicate a target RT60 to which we want
        to compare the estimated value.
    """

    h = np.array(h)
    fs = float(fs)

    # The power of the impulse response in dB
    power = h ** 2
    energy = np.cumsum(power[::-1])[::-1]  # Integration according to Schroeder

    # remove the possibly all zero tail
    i_nz = np.max(np.where(energy > 0)[0])
    energy = energy[:i_nz]
    energy_db = 10 * np.log10(energy)
    energy_db -= energy_db[0]

    # -5 dB headroom
    i_5db = np.min(np.where(-5 - energy_db > 0)[0])
    e_5db = energy_db[i_5db]
    t_5db = i_5db / fs

    # after decay
    i_decay = np.min(np.where(-5 - decay_db - energy_db > 0)[0])
    t_decay = i_decay / fs

    # compute the decay time
    decay_time = t_decay - t_5db
    est_rt60 = (60 / decay_db) * decay_time

    if plot:
        import matplotlib.pyplot as plt

        # Remove clip power below to minimum energy (for plotting purpose mostly)
        energy_min = energy[-1]
        energy_db_min = energy_db[-1]
        power[power < energy[-1]] = energy_min
        power_db = 10 * np.log10(power)
        power_db -= np.max(power_db)

        # time vector
        def get_time(x, fs):
            return np.arange(x.shape[0]) / fs - i_5db / fs

        T = get_time(power_db, fs)

        # plot power and energy
        plt.plot(get_time(energy_db, fs), energy_db, label="Energy")

        # now the linear fit
        plt.plot([0, est_rt60], [e_5db, -65], "--", label="Linear Fit")
        plt.plot(T, np.ones_like(T) * -60, "--", label="-60 dB")
        plt.vlines(
            est_rt60, energy_db_min, 0, linestyles="dashed", label="Estimated RT60"
        )

        if rt60_tgt is not None:
            plt.vlines(rt60_tgt, energy_db_min, 0, label="Target RT60")

        plt.legend()

        plt.show()

    return est_rt60



# Author: Eric Bezzam
# Date: July 15, 2016

import numpy as np
from .music import MUSIC
from scipy.linalg import svdvals

from scipy import linalg


class TOPS(MUSIC):
    """
    Class to apply Test of Orthogonality of Projected Subspaces [TOPS]_ for 
    Direction of Arrival (DoA) estimation.

    .. note:: Run locate_source() to apply the TOPS algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)

    References
    ----------

    .. [TOPS] Y. Yeo-Sun, L. M. Kaplan, J. H. McClellan, *TOPS: New DOA estimator for wideband 
        signals*, IEEE Trans. Signal Process., Vol. 54, Num 6., pp 1977--1989, 2006

    """

    def __init__(
        self,
        L,
        fs,
        nfft,
        c=343.0,
        num_src=1,
        mode="far",
        r=None,
        azimuth=None,
        colatitude=None,
        **kwargs
    ):

        MUSIC.__init__(
            self,
            L=L,
            fs=fs,
            nfft=nfft,
            c=c,
            num_src=num_src,
            mode=mode,
            r=r,
            azimuth=azimuth,
            colatitude=colatitude,
            **kwargs
        )

    def _process(self, X):
        """
        Perform TOPS for a given frame in order to estimate steered response 
        spectrum.
        """

        # need more than 1 frequency band
        if self.num_freq < 2:
            raise ValueError("Need more than one frequency band!")

        # select reference frequency (largest power)
        max_bin = np.argmax(
            np.sum(np.sum(abs(X[:, self.freq_bins, :]), axis=0), axis=1)
        )
        f0 = self.freq_bins[max_bin]
        freq = list(self.freq_bins)
        freq.remove(f0)

        # compute empirical cross correlation matrices
        C_hat = self._compute_correlation_matrices(X)

        # compute signal and noise subspace for each frequency band
        F = np.zeros((self.num_freq, self.M, self.num_src), dtype="complex64")
        W = np.zeros((self.num_freq, self.M, self.M - self.num_src), dtype="complex64")
        for k in range(self.num_freq):
            # subspace decomposition
            F[k, :, :], W[k, :, :], ws, wn = self._subspace_decomposition(
                C_hat[k, :, :]
            )

        # create transformation matrix
        f = (
            1.0
            / self.nfft
            / self.c
            * 1j
            * 2
            * np.pi
            * self.fs
            * (np.linspace(0, self.nfft // 2, self.nfft // 2 + 1) - f0)
        )

        Phi = np.zeros((len(f), self.M, self.grid.n_points), dtype="complex64")

        for n in range(self.grid.n_points):
            p_s = self.grid.cartesian[: self.grid.dim, n]
            proj = np.dot(p_s, self.L[: self.grid.dim, :])
            for m in range(self.M):
                Phi[:, m, n] = np.exp(f * proj[m])

        # determine direction using rank test
        for n in range(self.grid.n_points):
            # form D matrix
            D = np.zeros(
                (self.num_src, (self.M - self.num_src) * (self.num_freq - 1)),
                dtype="complex64",
            )
            for k in range(self.num_freq - 1):
                Uk = np.conjugate(np.dot(np.diag(Phi[k, :, n]), F[max_bin, :, :])).T
                # F[max_bin,:,:])).T
                idx = range(
                    k * (self.M - self.num_src), (k + 1) * (self.M - self.num_src)
                )
                D[:, idx] = np.dot(Uk, W[k, :, :])
            s = svdvals(D)
            self.grid.values[n] = 1.0 / s[-1]



# Author: Eric Bezzam
# Date: Feb 15, 2016
from __future__ import absolute_import, division, print_function

import math
import sys
import warnings
from abc import ABCMeta, abstractmethod

import numpy as np

from .grid import GridCircle, GridSphere
from .utils import polar_distance

"""Direction of Arrival (DoA) estimation."""


tol = 1e-14


class ModeVector(object):
    """
    This is a class for look-up tables of mode vectors. This look-up table
    is an outer product of three vectors running along candidate locations, time,
    and frequency. When the grid becomes large, the look-up table might be
    too large to store in memory. In that case, this class allows to only compute
    the outer product elements when needed, only keeping the three vectors in memory.
    When the table is small, a `precompute` option can be set to True to compute
    the whole table in advance.
    """

    def __init__(self, L, fs, nfft, c, grid, mode="far", precompute=False):
        """
        The constructor

        Parameters
        ----------
        L: ndarray
            contains the locations of the sensors in the columns of the array
        fs: int
            the sampling frequency of the input signal
        nfft: int
            the FFT length
        c: float
            the speed of sound
        grid: pyroomacoustcs.doa.Grid object
            the underlying grid on which to evaluate the mode vectors
        mode: string, optional
            specify if the mode vectors are far- or near-field
        precompute: bool
            if True, the whole look-up table is computed in advance
            (default False)
        """

        if nfft % 2 == 1:
            raise ValueError("Signal length must be even.")

        # this flag controls if the look-up table should be stored
        # or computed on the fly
        self.precompute = precompute

        # short hands for propagation vectors, upped to 3D array
        p_x = grid.x[None, None, :]
        p_y = grid.y[None, None, :]
        p_z = grid.z[None, None, :]

        # short hands for microphone locations, upped to 3D array
        r_x = L[0, None, :, None]
        r_y = L[1, None, :, None]

        if L.shape[0] == 3:
            r_z = L[2, None, :, None]
        else:
            r_z = np.zeros((1, L.shape[1], 1))

        # Here we compute the time of flights from source candidate locations
        # to microphones
        if mode == "near":
            # distance
            dist = np.sqrt((p_x - r_x) ** 2 + (p_y - r_y) ** 2 + (p_z - r_z) ** 2)

        elif mode == "far":
            # projection
            dist = (p_x * r_x) + (p_y * r_y) + (p_z * r_z)

        # shape (nfft // 2 + 1)
        self.tau = dist / c

        # shape (1, num_mics, grid_size)
        self.omega = 2 * np.pi * fs * np.arange(nfft // 2 + 1) / nfft

        if precompute:
            self.mode_vec = np.exp(1j * self.omega[:, None, None] * self.tau)
        else:
            self.mode_vec = None

    def __getitem__(self, ref):

        # If the look up table was precomputed
        if self.precompute:
            return self.mode_vec[ref]

        # we use this to test if an integer is passed
        integer = (
            int,
            np.int,
            np.int16,
            np.int32,
            np.int64,
            np.uint,
            np.uint16,
            np.uint32,
            np.uint64,
        )

        # Otherwise compute values on the fly
        if isinstance(ref[1], integer) and isinstance(ref[2], integer):
            w = self.omega[ref[0]]
        elif isinstance(ref[1], integer) or isinstance(ref[2], integer):
            w = self.omega[ref[0], None]
        else:
            w = self.omega[ref[0], None, None]

        if isinstance(ref[0], integer):
            tref0 = 0
        else:
            tref0 = slice(None, None, None)

        if len(ref) == 1:
            return np.exp(1j * w * self.tau[tref0, :, :])
        elif len(ref) == 2:
            return np.exp(1j * w * self.tau[tref0, ref[1], :])
        elif len(ref) == 3:
            return np.exp(1j * w * self.tau[tref0, ref[1], ref[2]])
        else:
            raise ValueError("Too many axis")


class DOA(object):
    """

    Abstract parent class for Direction of Arrival (DoA) algorithms. After 
    creating an object (SRP, MUSIC, CSSM, WAVES, or TOPS), run locate_source to
    apply the corresponding algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    n_grid: int
        If azimuth and colatitude are not specified, we will create a grid with
        so many points. Default is 360.
    dim: int
        The dimension of the problem. Set dim=2 to find sources on the circle (x-y plane).
        Set dim=3 to search on the whole sphere.
    """

    __metaclass__ = ABCMeta

    def __init__(
        self,
        L,
        fs,
        nfft,
        c=343.0,
        num_src=1,
        mode="far",
        r=None,
        azimuth=None,
        colatitude=None,
        n_grid=None,
        dim=2,
        *args,
        **kwargs
    ):

        if dim > L.shape[0]:
            raise ValueError("Microphones locations missing dimensions.")

        self.L = L  # locations of mics
        self.fs = fs  # sampling frequency
        self.c = c  # speed of sound
        self.M = L.shape[1]  # number of microphones
        self.D = L.shape[0]  # number of dimensions (x,y,z)
        self.num_snap = None  # number of snapshots

        self.nfft = nfft
        self.max_bin = int(self.nfft // 2) + 1
        self.freq_bins = None
        self.freq_hz = None
        self.num_freq = None

        self.num_src = self._check_num_src(num_src)
        self.sources = np.zeros([self.D, self.num_src])
        self.src_idx = np.zeros(self.num_src, dtype=np.int)
        self.azimuth_recon = None
        self.colatitude_recon = None
        self.alpha_recon = None

        self.mode = mode
        if self.mode == "far":
            self.r = np.ones(1)
        elif r is None:
            self.r = np.ones(1)
            self.mode = "far"
        else:
            self.r = r
            if r == np.ones(1):
                mode = "far"

        # Set the dimension of the problem
        if dim != 2 and dim != 3:
            raise ValueError("The dimension must be 2 or 3.")
        else:
            self.dim = dim

        # Some logic to create a grid for search
        self.grid = None
        if azimuth is None and colatitude is None:

            # Use a default grid size
            if dim == 2:

                if n_grid is None:
                    n_grid = 360

                self.grid = GridCircle(n_points=n_grid)

            elif dim == 3:

                if n_grid is None:
                    n_grid = 180 * 90

                self.grid = GridSphere(n_points=n_grid)

        elif azimuth is None and colatitude is not None:

            raise ValueError("Azimuth should always be specified.")

        else:

            if dim == 2:

                if colatitude is not None:
                    warnings.warn("Colatitude is ignored for 2D problems.")

                self.grid = GridCircle(azimuth=azimuth)

            elif dim == 3:

                if azimuth.ndim != 1:
                    raise ValueError("Azimuth should be a 1D ndarray.")

                if colatitude is None:
                    warnings.warn(
                        "Colatitude is not specified. Setting all colatitude to pi / 2."
                    )

                    colatitude = (np.pi / 2) * np.ones(azimuth.shape[0])
                    grid_points = np.vstack((azimuth, colatitude))

                    self.grid = GridSphere(spherical_points=grid_points)

                else:

                    # when both azimuth and theta are specified,
                    # we assume we want the cartesian product
                    A, C = np.meshgrid(np.unique(azimuth), np.unique(colatitude))
                    grid_points = np.vstack((A.flatten(), C.flatten()))

                    self.grid = GridSphere(spherical_points=grid_points)

        # spatial spectrum / dirty image (FRIDA)
        self.P = None

        # build lookup table to candidate locations from r, azimuth, colatitude
        from .frida import FRIDA

        if not isinstance(self, FRIDA):
            self.mode_vec = ModeVector(self.L, self.fs, self.nfft, self.c, self.grid)

    def locate_sources(
        self, X, num_src=None, freq_range=[500.0, 4000.0], freq_bins=None, freq_hz=None
    ):
        """
        Locate source(s) using corresponding algorithm.

        Parameters
        ----------
        X: numpy array
            Set of signals in the frequency (RFFT) domain for current 
            frame. Size should be M x F x S, where M should correspond to the 
            number of microphones, F to nfft/2+1, and S to the number of snapshots 
            (user-defined). It is recommended to have S >> M.
        num_src: int
            Number of sources to detect. Default is value given to 
            object constructor.
        freq_range: list of floats, length 2
            Frequency range on which to run DoA: [fmin, fmax].
        freq_bins: list of int
            freq_bins: List of individual frequency bins on which to run 
            DoA. 
            If defined by user, it will **not** take into consideration freq_range 
            or freq_hz.
        freq_hz: list of floats
            List of individual frequencies on which to run DoA. If 
            defined by user, it will **not** take into consideration freq_range.
        """
        # check validity of inputs
        if num_src is not None and num_src != self.num_src:
            self.num_src = self._check_num_src(num_src)
            self.sources = np.zeros([self.num_src, self.D])
            self.src_idx = np.zeros(self.num_src, dtype=np.int)
            self.angle_of_arrival = None
        if X.shape[0] != self.M:
            raise ValueError(
                "Number of signals (rows) does not match the \
                number of microphones."
            )
        if X.shape[1] != self.max_bin:
            raise ValueError("Mismatch in FFT length.")
        self.num_snap = X.shape[2]

        # frequency bins on which to apply DOA
        if freq_bins is not None:
            self.freq_bins = np.array(freq_bins, dtype=np.int)
        elif freq_hz is not None:
            self.freq_bins = [int(np.round(f / self.fs * self.nfft)) for f in freq_bins]
        else:
            freq_range = [int(np.round(f / self.fs * self.nfft)) for f in freq_range]
            self.freq_bins = np.arange(freq_range[0], freq_range[1], dtype=np.int)

        self.freq_bins = self.freq_bins[self.freq_bins < self.max_bin]
        self.freq_bins = self.freq_bins[self.freq_bins >= 0]
        self.freq_hz = self.freq_bins * float(self.fs) / float(self.nfft)
        self.num_freq = len(self.freq_bins)

        # search for DoA according to desired algorithm

        # initialize the grid value to zero
        self.grid.set_values(0.0)

        # Run the algorithm
        self._process(X)

        # locate sources
        from .frida import FRIDA

        if not isinstance(self, FRIDA):

            self.src_idx = self.grid.find_peaks(k=self.num_src)

            self.num_src = len(self.src_idx)

            if self.dim == 2:
                self.azimuth_recon = self.grid.azimuth[self.src_idx]
            elif self.dim == 3:
                self.azimuth_recon = self.grid.azimuth[self.src_idx]
                self.colatitude_recon = self.grid.colatitude[self.src_idx]

    def polar_plt_dirac(
        self,
        azimuth_ref=None,
        alpha_ref=None,
        save_fig=False,
        file_name=None,
        plt_dirty_img=True,
    ):
        """
        Generate polar plot of DoA results.

        Parameters
        ----------
        azimuth_ref: numpy array
            True direction of sources (in radians).
        alpha_ref: numpy array
            Estimated amplitude of sources.
        save_fig: bool
            Whether or not to save figure as pdf.
        file_name: str
            Name of file (if saved). Default is 
            'polar_recon_dirac.pdf'
        plt_dirty_img: bool
            Whether or not to plot spatial spectrum or 
            'dirty image' in the case of FRI.
        """

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        if self.dim != 2:
            raise ValueError("This function only handles 2D problems.")

        azimuth_recon = self.azimuth_recon
        num_mic = self.M
        phi_plt = self.grid.azimuth

        # determine amplitudes
        from .frida import FRIDA

        if not isinstance(self, FRIDA):  # use spatial spectrum

            dirty_img = self.grid.values
            alpha_recon = self.grid.values[self.src_idx]
            alpha_ref = alpha_recon

        else:  # create dirty image

            dirty_img = self._gen_dirty_img()
            alpha_recon = np.mean(np.abs(self.alpha_recon), axis=1)
            alpha_recon /= alpha_recon.max()
            if alpha_ref is None:  # non-simulated case
                alpha_ref = alpha_recon

        # plot
        fig = plt.figure(figsize=(5, 4), dpi=90)
        ax = fig.add_subplot(111, projection="polar")
        base = 1.0
        height = 10.0
        blue = [0, 0.447, 0.741]
        red = [0.850, 0.325, 0.098]

        if azimuth_ref is not None:
            if alpha_ref.shape[0] < azimuth_ref.shape[0]:
                alpha_ref = np.concatenate(
                    (alpha_ref, np.zeros(azimuth_ref.shape[0] - alpha_ref.shape[0]))
                )

            # match detected with truth
            recon_err, sort_idx = polar_distance(azimuth_recon, azimuth_ref)
            if self.num_src > 1:
                azimuth_recon = azimuth_recon[sort_idx[:, 0]]
                alpha_recon = alpha_recon[sort_idx[:, 0]]
                azimuth_ref = azimuth_ref[sort_idx[:, 1]]
                alpha_ref = alpha_ref[
                    sort_idx[:, 0]
                ]  # Robin: not sure why index 0 works here...
            elif azimuth_ref.shape[0] > 1:  # one detected source
                alpha_ref[sort_idx[1]] = alpha_recon

            # markers for original doa
            K = len(azimuth_ref)
            ax.scatter(
                azimuth_ref,
                base + height * alpha_ref,
                c=np.tile(blue, (K, 1)),
                s=70,
                alpha=0.75,
                marker="^",
                linewidths=0,
                label="original",
            )

            # stem for original doa
            if K > 1:
                for k in range(K):
                    ax.plot(
                        [azimuth_ref[k], azimuth_ref[k]],
                        [base, base + height * alpha_ref[k]],
                        linewidth=1.5,
                        linestyle="-",
                        color=blue,
                        alpha=0.6,
                    )
            else:
                ax.plot(
                    [azimuth_ref, azimuth_ref],
                    [base, base + height * alpha_ref],
                    linewidth=1.5,
                    linestyle="-",
                    color=blue,
                    alpha=0.6,
                )

        K_est = azimuth_recon.size

        # markers for reconstructed doa
        ax.scatter(
            azimuth_recon,
            base + height * alpha_recon,
            c=np.tile(red, (K_est, 1)),
            s=100,
            alpha=0.75,
            marker="*",
            linewidths=0,
            label="reconstruction",
        )

        # stem for reconstructed doa
        if K_est > 1:
            for k in range(K_est):
                ax.plot(
                    [azimuth_recon[k], azimuth_recon[k]],
                    [base, base + height * alpha_recon[k]],
                    linewidth=1.5,
                    linestyle="-",
                    color=red,
                    alpha=0.6,
                )

        else:
            ax.plot(
                [azimuth_recon, azimuth_recon],
                [base, base + height * alpha_recon],
                linewidth=1.5,
                linestyle="-",
                color=red,
                alpha=0.6,
            )

            # plot the 'dirty' image
        if plt_dirty_img:
            dirty_img = np.abs(dirty_img)
            min_val = dirty_img.min()
            max_val = dirty_img.max()
            dirty_img = (dirty_img - min_val) / (max_val - min_val)

            # we need to make a complete loop, copy first value to last
            c_phi_plt = np.r_[phi_plt, phi_plt[0]]
            c_dirty_img = np.r_[dirty_img, dirty_img[0]]
            ax.plot(
                c_phi_plt,
                base + height * c_dirty_img,
                linewidth=1,
                alpha=0.55,
                linestyle="-",
                color=[0.466, 0.674, 0.188],
                label="spatial spectrum",
            )

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(
            handles=handles[:3],
            framealpha=0.5,
            scatterpoints=1,
            loc=8,
            fontsize=9,
            ncol=1,
            bbox_to_anchor=(0.9, -0.17),
            handletextpad=0.2,
            columnspacing=1.7,
            labelspacing=0.1,
        )

        ax.set_xlabel(r"azimuth ${\varphi}$", fontsize=11)
        ax.set_xticks(np.linspace(0, 2 * np.pi, num=12, endpoint=False))
        ax.xaxis.set_label_coords(0.5, -0.11)
        ax.set_yticks(np.linspace(0, 1, 2))
        ax.xaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle=":")
        ax.yaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle="--")
        ax.set_ylim([0, 1.05 * (base + height)])

        plt.tight_layout()

        if save_fig:
            if file_name is None:
                file_name = "polar_recon_dirac.pdf"
            plt.savefig(file_name, format="pdf", dpi=300, transparent=True)

    def _check_num_src(self, num_src):
        # # check validity of inputs
        # if num_src > self.M:
        #     warnings.warn('Number of sources cannot be more than number of \
        #         microphones. Changing number of sources to ' +
        #         str(self.M) + '.')
        #     num_src = self.M
        if num_src < 1:
            warnings.warn(
                "Number of sources must be at least 1. Changing \
                number of sources to 1."
            )
            num_src = 1
        valid = int(num_src)
        return valid



# Test of Wall class and intersection methods
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import unittest
import numpy as np

import pyroomacoustics as pra


c0 = np.array([[0, 3, 3, 0], [0, 0, 0, 0], [0, 0, 2, 2],])  # left
c1 = np.array([[0, 0, 6, 6], [8, 8, 8, 8], [0, 4, 4, 0],])  # right
c2 = np.array([[0, 0, 6, 3,], [0, 8, 8, 0,], [0, 0, 0, 0,],])  # floor
c3 = np.array([[0, 3, 6, 0,], [0, 0, 8, 8,], [2, 2, 4, 4,],])  # ceiling
c4 = np.array([[0, 0, 0, 0,], [0, 0, 8, 8,], [0, 2, 4, 0,],])  # back
c5 = np.array([[3, 6, 6, 3,], [0, 8, 8, 0,], [0, 0, 4, 2,],])  # front

# Strange room (not square and inclined roof)
wall_corners_3D = [c0, c1, c2, c3, c4, c5]

absorptions_3D = [0.1, 0.25, 0.25, 0.25, 0.2, 0.15]
scatterings_3D = [0.5, 0.1, 0.4, 0.2, 0.3, 0.5]


# Let's describe a pentagonal room with corners :
# (-1,0) (-1,2) (0,3) (2,2) (2,-1)
d0 = np.array([[-1, -1], [0, 2],])  # side1
d1 = np.array([[-1, 0], [2, 3],])  # side2
d2 = np.array([[0, 2], [3, 2],])  # side3
d3 = np.array([[2, 2], [2, -1],])  # side4
d4 = np.array([[2, -1], [-1, 0],])  # side5
wall_corners_2D = [d0, d1, d2, d3, d4]


# Let's describe a simple non_convex room with corners :
# (0,0) (0,2) (1,1) (2,2) (2,0)
e0 = np.array([[0, 0], [0, 2],])  # side1
e1 = np.array([[0, 1], [2, 1],])  # side2
e2 = np.array([[1, 2], [1, 2],])  # side3
e3 = np.array([[2, 2], [2, 0],])  # side4
e4 = np.array([[2, 0], [0, 0],])  # side5

wall_corners_2D_non_convex = [e0, e1, e2, e3, e4]
absorptions_2D = [0.1, 0.1, 0.1, 0.1, 0.1]
scatterings_2D = [0.1, 0.2, 0.3, 0.4, 0.5]


f0 = np.array([[0, 0], [0, 3],])  # side1
f1 = np.array([[0, 4], [3, 3],])  # side2
f2 = np.array([[4, 4], [3, 0],])  # side3
f3 = np.array([[4, 0], [0, 0],])  # side4

wall_corners_2D_shoebox = [f0, f1, f2, f3]
absorptions_shoebox = [0.1, 0.1, 0.1, 0.1]


def room_factory(walls, obstructing_walls, microphones):

    args = [
        walls,
        obstructing_walls,
        [],
        pra.constants.get("c"),  # speed of sound
        3,
        1e-7,  # energy_thres
        1.0,  # time_thres
        0.5,  # receiver_radius
        0.004,  # hist_bin_size
        True,  # a priori we will always use a hybrid model
    ]

    if walls[0].dim == 2:
        room = pra.libroom.Room2D(*args)
    else:
        room = pra.libroom.Room(*args)

    for m in range(microphones.shape[1]):
        room.add_mic(microphones[:, None, m])

    return room


class TestRoomWalls(unittest.TestCase):
    def test_max_dist_3D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_3D, absorptions_3D, scatterings_2D)
        ]
        obstructing_walls = []
        microphones = np.array([[1,], [1,], [1,],])

        room = room_factory(walls, obstructing_walls, microphones)

        eps = 0.001
        result = room.get_max_distance()
        correct = np.sqrt(116) + 1
        self.assertTrue(abs(result - correct) < eps)

    def test_max_dist_2D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_2D, absorptions_2D, scatterings_2D)
        ]
        obstructing_walls = []
        microphones = np.array([[1,], [1,],])

        room = room_factory(walls, obstructing_walls, microphones)

        eps = 0.001
        result = room.get_max_distance()
        self.assertEqual(result, np.sqrt(25) + 1)

    def test_same_wall_true3D(self):
        w1 = pra.wall_factory(
            wall_corners_3D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        w2 = pra.wall_factory(
            wall_corners_3D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        self.assertTrue(w1.same_as(w2))

    def test_same_wall_true2D(self):
        w1 = pra.wall_factory(
            wall_corners_2D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        w2 = pra.wall_factory(
            wall_corners_2D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        self.assertTrue(w1.same_as(w2))

    def test_same_wall_false3D(self):
        w1 = pra.wall_factory(
            wall_corners_3D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        w2 = pra.wall_factory(
            wall_corners_3D[1], absorptions_3D[:1], scatterings_3D[:1]
        )
        self.assertTrue(not w1.same_as(w2))

    def test_same_wall_false3D_more_corners(self):

        # Modification of wall_corners_3D[0]: adding a corner => 5 corners wall
        c1 = np.array([[0, 3, 3, 1.5, 0], [0, 0, 0, 0, 0], [0, 0, 2, 1.5, 2]])

        w1 = pra.wall_factory(
            wall_corners_3D[0], absorptions_3D[:1], scatterings_3D[:1]
        )
        w2 = pra.wall_factory(c1, absorptions_3D[:1], scatterings_3D[:1])
        self.assertTrue(not w1.same_as(w2))

    def test_next_wall_hit(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_3D, absorptions_3D, scatterings_3D)
        ]
        obstructing_walls = []
        microphones = np.array([[1.0,], [1.0,], [1.0,],])

        room = room_factory(walls, obstructing_walls, microphones)

        eps = 0.001

        # start on one empty space
        start = [-2, 4, 1]

        # end at the same (x,y) but very high in the sky
        end = [5, -3, 1]

        # correct intersection
        result = np.array([0, 2, 1])

        ttuple = np.array(room.next_wall_hit(start, end, False))

        correct_result = np.allclose(ttuple[0], result, atol=eps)
        correct_next_wall = ttuple[1] == 4
        correct_distance = np.allclose(ttuple[2], np.linalg.norm(result - start))

        self.assertTrue(correct_next_wall and correct_result and correct_distance)

    def test_next_wall_nohit(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_3D, absorptions_3D, scatterings_3D)
        ]
        obstructing_walls = []
        microphones = np.array([[1,], [1,], [1,],])

        room = room_factory(walls, obstructing_walls, microphones)

        eps = 0.001

        # start outside the room
        start = [-1, -1, -1]

        # end outside the room
        end = [-2, -3, -1]

        ttuple = np.array(room.next_wall_hit(start, end, False))

        self.assertTrue(ttuple[1] == -1)

    def test_next_wall_hit2D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_2D, absorptions_2D, scatterings_2D)
        ]
        obstructing_walls = []
        microphones = np.array([[1.0,], [1.0,],])

        room = room_factory(walls, obstructing_walls, microphones)

        eps = 0.001

        # start on one empty space
        start = [0, -2]

        # end at the same (x,y) but very high in the sky
        end = [0, 4]

        # the intersection expected result
        result = np.array([0, -1.0 / 3])

        ttuple = np.array(room.next_wall_hit(start, end, False))

        correct_result = np.allclose(ttuple[0], result, atol=eps)
        correct_next_wall = ttuple[1] == 4
        correct_distance = np.allclose(ttuple[2], np.linalg.norm(result - start))

        self.assertTrue(correct_next_wall and correct_result and correct_distance)

    def test_scat_ray_blocked(self):
        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(
                wall_corners_2D_non_convex, absorptions_2D, scatterings_2D
            )
        ]
        obstructing_walls = [1, 2]  # index of the 2 possibly obstructing walls
        microphones = np.array([[1.5], [1.2]])

        room = room_factory(walls, obstructing_walls, microphones)
        room.set_params(
            pra.constants.get("c"),
            2,  # order of ISM
            0.001,  # energy threshold for rays
            200.0,  # time threshold for rays
            0.1,  # detector radius
            0.004,  # histogram time resolution
            True,  # hybrid model
        )

        prev_wall = room.get_wall(0)

        prev_last_hit = [0.5, 0.0]
        last_hit = [0, 1.9]
        total_dist = 0.0

        energy = [1000000.0]
        scatter_coef = scatterings_2D[0]

        output = [
            [pra.libroom.Hit(1)]
        ]  # arbitrary initialisation to have the correct shape
        self.assertTrue(
            not room.scat_ray(energy, prev_wall, prev_last_hit, last_hit, total_dist)
        )

    def test_scat_ray_ok(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(
                wall_corners_2D_non_convex, absorptions_2D, scatterings_2D
            )
        ]
        obstructing_walls = [1, 2]
        microphones = np.array([[0.5], [0.2]])

        room = room_factory(walls, obstructing_walls, microphones)
        room.set_params(
            pra.constants.get("c"),
            2,  # order of ISM
            0.001,  # energy threshold for rays
            200.0,  # time threshold for rays
            0.1,  # detector radius
            0.004,  # histogram time resolution
            True,  # hybrid model
        )

        prev_wall = room.get_wall(0)

        prev_last_hit = [2, 0.2]
        last_hit = [0, 1.9]
        total_dist = 0.0

        energy = [1000000.0]
        scatter_coef = scatterings_2D[0]

        output = [
            [pra.libroom.Hit(1)]
        ]  # arbitrary initialisation to have the correct shape
        self.assertTrue(
            room.scat_ray(energy, prev_wall, prev_last_hit, last_hit, total_dist)
        )

    """
    def test_scat_ray_energy(self):

        # Test energy with the energy / (4*pi*dist) rule

        walls = [pra.wall_factory(c, [a], [s]) for c, a, s in zip(wall_corners_2D_shoebox, absorptions_shoebox, scatterings_2D)]
        obstructing_walls = []
        microphones = np.array([
            [3.1],
            [1.5 ]
            ])

        mic_pos = microphones[:, 0]
        print(mic_pos)

        room = room_factory(walls, obstructing_walls, microphones)
        room.set_params(
                pra.constants.get('c'),
                2,  # order of ISM
                1e-7,  # energy threshold for rays
                5.,  # time threshold for rays
                0.1,  # detector radius
                0.004,  # histogram time resolution
                True,  # hybrid model
                )

        prev_wall = room.get_wall(0)

        total_dist = 2. #total to reach last hit
        prev_last_hit = [2,1.5]
        last_hit = [0, 1.5]

        energy = [1.]
        scatter_coef = 0.1

        eps = 0.0001
        result = room.compute_scat_energy(energy, scatter_coef, prev_wall, prev_last_hit, last_hit, mic_pos, total_dist)
        self.assertTrue(np.allclose(result, np.sqrt(0.1*2*(1-np.sqrt(3.1*3.1-0.1*0.1)/3.1))/(5), atol=eps))
    """

    def test_contains_2D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(
                wall_corners_2D_non_convex, absorptions_2D, scatterings_2D
            )
        ]
        obstructing_walls = []
        microphones = np.array([[0.5], [0.2]])

        room = room_factory(walls, obstructing_walls, microphones)

        self.assertTrue(room.contains(microphones[:, 0]))

    def test_notcontains_2D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(
                wall_corners_2D_non_convex, absorptions_2D, scatterings_2D
            )
        ]
        obstructing_walls = []
        microphones = np.array([[1.0], [1.7]])

        room = room_factory(walls, obstructing_walls, microphones)

        self.assertTrue(not room.contains(microphones[:, 0]))

    def test_contains_3D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_3D, absorptions_3D, scatterings_3D)
        ]
        obstructing_walls = []
        microphones = np.array([[1.0], [1.0], [1.0]])

        room = room_factory(walls, obstructing_walls, microphones)

        self.assertTrue(room.contains(microphones[:, 0]))

    def test_notcontains_3D(self):

        walls = [
            pra.wall_factory(c, [a], [s])
            for c, a, s in zip(wall_corners_3D, absorptions_3D, scatterings_3D)
        ]
        obstructing_walls = []
        microphones = np.array([[5.0], [7.0], [40]])

        room = room_factory(walls, obstructing_walls, microphones)

        self.assertTrue(not room.contains(microphones[:, 0]))


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pyroomacoustics as pra
from scipy.io import wavfile
import unittest

np.random.seed(0)

# We use several sound samples for each source to have a long enough length
wav_files = [
    [
        "examples/input_samples/cmu_arctic_us_axb_a0004.wav",
        "examples/input_samples/cmu_arctic_us_axb_a0005.wav",
        "examples/input_samples/cmu_arctic_us_axb_a0006.wav",
    ],
    [
        "examples/input_samples/cmu_arctic_us_aew_a0001.wav",
        "examples/input_samples/cmu_arctic_us_aew_a0002.wav",
        "examples/input_samples/cmu_arctic_us_aew_a0003.wav",
    ],
]

# List of frame lengths to test
L = [256, 512, 1024, 2048, 4096]

# Frequency Blind Source Separation
def freq_bss(algo="auxiva", L=256, **kwargs):

    # Room dimensions in meters
    room_dim = [8, 9]

    # create a room with sources and mics
    room = pra.ShoeBox(room_dim, fs=16000, max_order=0, sigma2_awgn=1e-8)

    # get signals
    signals = [
        np.concatenate([wavfile.read(f)[1].astype(np.float32) for f in source_files])
        for source_files in wav_files
    ]
    delays = [1.0, 0.0]
    if algo == "overiva":
        locations = [[2.5, 3]]
    else:
        locations = [[2.5, 3], [2.5, 6]]

    # add mic and good source to room
    # Add silent signals to all sources
    for sig, d, loc in zip(signals, delays, locations):
        room.add_source(loc, signal=np.zeros_like(sig), delay=d)

    # add microphone array
    room.add_microphone_array(
        pra.MicrophoneArray(np.c_[[6.5, 4.49], [6.5, 4.51]], fs=room.fs)
    )

    # compute RIRs
    room.compute_rir()

    # Record each source separately
    separate_recordings = []
    for source, signal in zip(room.sources, signals):

        source.signal[:] = signal

        room.simulate()
        separate_recordings.append(room.mic_array.signals)

        source.signal[:] = 0.0
    separate_recordings = np.array(separate_recordings)

    # Mix down the recorded signals
    mics_signals = np.sum(separate_recordings, axis=0)

    ## STFT analysis
    # shape == (n_chan, n_frames, n_freq)
    X = pra.transform.stft.analysis(
        mics_signals.T, L, L, zp_front=L // 2, zp_back=L // 2
    )

    ## START BSS
    if algo == "auxiva":
        # Run AuxIVA
        Y = pra.bss.auxiva(X, n_iter=30, proj_back=True, **kwargs)
        max_mse = 5e-2
    elif algo == "ilrma":
        # Run ILRMA
        Y = pra.bss.ilrma(X, n_iter=30, n_components=2, proj_back=True, **kwargs)
        max_mse = 5e-2
    elif algo == "sparseauxiva":
        # Estimate set of active frequency bins
        ratio = 0.35
        average = np.abs(np.mean(np.mean(X, axis=2), axis=0))
        k = np.int_(average.shape[0] * ratio)
        S = np.sort(np.argpartition(average, -k)[-k:])
        # Run SparseAuxIva
        Y = pra.bss.sparseauxiva(X, S, n_iter=30, proj_back=True, **kwargs)
        max_mse = 1.5e-1
    elif algo == "overiva":
        Y = pra.bss.auxiva(X, n_src=1, n_iter=30, proj_back=True, **kwargs)
        max_mse = 0.5
    elif algo == "fastmnmf":
        Y = pra.bss.fastmnmf(X, n_src=2, n_iter=30, n_components=8)
        max_mse = 1e-1

    ## STFT Synthesis
    if algo == "overiva":
        y = pra.transform.stft.synthesis(
            Y[:, :, 0], L, L, zp_front=L // 2, zp_back=L // 2
        ).T
        y = y[None, :]
    else:
        y = pra.transform.stft.synthesis(Y, L, L, zp_front=L // 2, zp_back=L // 2).T

    # Calculate MES
    #############
    ref = np.moveaxis(separate_recordings, 1, 2)
    y_aligned = y[:, L // 2 : ref.shape[1] + L // 2]

    mse = np.mean((ref[:, : y_aligned.shape[1], 0] - y_aligned) ** 2)
    ref_var = np.var(np.concatenate(ref[:, : y_aligned.shape[1], 0]))

    print(
        "%s with a %d frame length: Relative MSE (expected less than %.e)"
        % (algo, L, max_mse),
        mse / ref_var,
    )
    assert (mse / ref_var) < max_mse

    # Now test other parameter combinations, just run, no output check


class TestBSS(unittest.TestCase):
    # Test auxiva with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_auxiva_laplace(self):
        for block in L:
            freq_bss(algo="auxiva", L=block, model="laplace")

    # Test auxiva with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_auxiva_gauss(self):
        for block in L:
            freq_bss(algo="auxiva", L=block, model="gauss")

    # Test ilrma with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_ilrma(self):
        for block in L:
            freq_bss(algo="ilrma", L=block)

    # Test sparse auxiva with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_sparse_auxiva_laplace(self):
        for block in L:
            freq_bss(algo="sparseauxiva", L=block, model="laplace")

    # Test sparse auxiva with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_sparse_auxiva_gauss(self):
        for block in L:
            freq_bss(algo="sparseauxiva", L=block, model="gauss")

    # Test overiva with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_overiva(self):
        for block in L:
            freq_bss(algo="overiva", L=block)

    # Test fastmnmf with frame lengths [256, 512, 1024, 2048, 4096]
    def test_bss_fastmnmf(self):
        for block in L:
            freq_bss(algo="fastmnmf", L=block)


if __name__ == "__main__":
    unittest.main()



import numpy as np

import pyroomacoustics as pra

eps = 1e-10
room_dim = [10, 7.5, 3.5]  # meters
fs = 16000

V = np.prod(room_dim)
S = 2 * (
    room_dim[0] * room_dim[1] + room_dim[1] * room_dim[2] + room_dim[2] * room_dim[0]
)


def test_rt60_theory_single_band():

    # The desired reverberation time and dimensions of the room
    rt60_tgt = 0.3  # seconds

    # We invert Sabine's formula to obtain the parameters for the ISM simulator
    e_absorption, max_order = pra.inverse_sabine(rt60_tgt, room_dim)

    # Create the room
    room = pra.ShoeBox(
        room_dim, fs=fs, materials=pra.Material(e_absorption), max_order=max_order
    )

    rt60_sabine = pra.rt60_sabine(S, V, e_absorption, 0.0, room.c)
    assert (rt60_sabine - room.rt60_theory(formula="sabine")) < eps

    rt60_eyring = pra.rt60_eyring(S, V, e_absorption, 0.0, room.c)
    assert (rt60_eyring - room.rt60_theory(formula="eyring")) < eps


def test_rt60_theory_multi_band():

    # Create the room
    room = pra.ShoeBox(room_dim, fs=fs, materials=pra.Material("curtains_cotton_0.5"),)

    # run the different rt60 functions
    room.rt60_theory(formula="sabine")
    room.rt60_theory(formula="eyring")


def test_rt60_measure():

    # Create the room
    room = pra.ShoeBox(
        room_dim, fs=fs, materials=pra.Material("curtains_cotton_0.5"), max_order=10,
    )

    # place the source in the room
    room.add_source([2.5, 3.73, 1.76])

    # place a microphone in the room
    room.add_microphone([6.3, 4.87, 1.2])

    room.compute_rir()

    room.measure_rt60()


if __name__ == "__main__":

    test_rt60_theory_single_band()
    test_rt60_theory_multi_band()
    test_rt60_measure()



'''
A collection of functions to plot maps and points on circles and spheres.
'''
import numpy as np

def polar_plt_dirac(self, azimuth_ref=None, alpha_ref=None, save_fig=False, 
        file_name=None, plt_dirty_img=True):
    """
    Generate polar plot of DoA results.

    Parameters
    ----------
    azimuth_ref: numpy array
        True direction of sources (in radians).
    alpha_ref: numpy array
        Estimated amplitude of sources.
    save_fig: bool
        Whether or not to save figure as pdf.
    file_name: str
        Name of file (if saved). Default is 
        'polar_recon_dirac.pdf'
    plt_dirty_img: bool
        Whether or not to plot spatial spectrum or 
        'dirty image' in the case of FRI.
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings
        warnings.warn('Matplotlib is required for plotting')
        return

    if self.dim != 2:
        raise ValueError('This function only handles 2D problems.')

    azimuth_recon = self.azimuth_recon
    num_mic = self.M
    phi_plt = self.grid.azimuth

    # determine amplitudes
    from fri import FRI
    if not isinstance(self, FRI):  # use spatial spectrum

        dirty_img = self.grid.values
        alpha_recon = self.grid.values[self.src_idx]
        alpha_ref = alpha_recon

    else:  # create dirty image

        dirty_img = self._gen_dirty_img()
        alpha_recon = np.mean(np.abs(self.alpha_recon), axis=1)
        alpha_recon /= alpha_recon.max()
        if alpha_ref is None:   # non-simulated case
            alpha_ref = alpha_recon

    # plot
    fig = plt.figure(figsize=(5, 4), dpi=90)
    ax = fig.add_subplot(111, projection='polar')
    base = 1.
    height = 10.
    blue = [0, 0.447, 0.741]
    red = [0.850, 0.325, 0.098]

    if azimuth_ref is not None:
        if alpha_ref.shape[0] < azimuth_ref.shape[0]:
            alpha_ref = np.concatenate((alpha_ref,np.zeros(azimuth_ref.shape[0]-
                alpha_ref.shape[0])))

        # match detected with truth
        recon_err, sort_idx = polar_distance(azimuth_recon, azimuth_ref)
        if self.num_src > 1:
            azimuth_recon = azimuth_recon[sort_idx[:, 0]]
            alpha_recon = alpha_recon[sort_idx[:, 0]]
            azimuth_ref = azimuth_ref[sort_idx[:, 1]]
            alpha_ref = alpha_ref[sort_idx[:, 1]]
        elif azimuth_ref.shape[0] > 1:   # one detected source
            alpha_ref[sort_idx[1]] =  alpha_recon

        # markers for original doa
        K = len(azimuth_ref)
        ax.scatter(azimuth_ref, base + height*alpha_ref, c=np.tile(blue, 
            (K, 1)), s=70, alpha=0.75, marker='^', linewidths=0, 
            label='original')

        # stem for original doa
        if K > 1:
            for k in range(K):
                ax.plot([azimuth_ref[k], azimuth_ref[k]], [base, base + 
                    height*alpha_ref[k]], linewidth=1.5, linestyle='-', 
                    color=blue, alpha=0.6)
        else:
            ax.plot([azimuth_ref, azimuth_ref], [base, base + height*alpha_ref], 
                linewidth=1.5, linestyle='-', color=blue, alpha=0.6)


    K_est = azimuth_recon.size

    # markers for reconstructed doa
    ax.scatter(azimuth_recon, base + height*alpha_recon, c=np.tile(red, 
        (K_est, 1)), s=100, alpha=0.75, marker='*', linewidths=0, 
        label='reconstruction')

    # stem for reconstructed doa
    if K_est > 1:
        for k in range(K_est):
            ax.plot([azimuth_recon[k], azimuth_recon[k]], [base, base + 
                height*alpha_recon[k]], linewidth=1.5, linestyle='-', 
                color=red, alpha=0.6)

    else:
        ax.plot([azimuth_recon, azimuth_recon], [base, base + height*alpha_recon], 
            linewidth=1.5, linestyle='-', color=red, alpha=0.6)            

    # plot the 'dirty' image
    if plt_dirty_img:
        dirty_img = np.abs(dirty_img)
        min_val = dirty_img.min()
        max_val = dirty_img.max()
        dirty_img = (dirty_img - min_val) / (max_val - min_val)

        # we need to make a complete loop, copy first value to last
        c_phi_plt = np.r_[phi_plt, phi_plt[0]]
        c_dirty_img = np.r_[dirty_img, dirty_img[0]]
        ax.plot(c_phi_plt, base + height*c_dirty_img, linewidth=1, 
            alpha=0.55,linestyle='-', color=[0.466, 0.674, 0.188], 
            label='spatial spectrum')

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles=handles[:3], framealpha=0.5,
              scatterpoints=1, loc=8, fontsize=9,
              ncol=1, bbox_to_anchor=(0.9, -0.17),
              handletextpad=.2, columnspacing=1.7, labelspacing=0.1)

    ax.set_xlabel(r'azimuth $\bm{\varphi}$', fontsize=11)
    ax.set_xticks(np.linspace(0, 2 * np.pi, num=12, endpoint=False))
    ax.xaxis.set_label_coords(0.5, -0.11)
    ax.set_yticks(np.linspace(0, 1, 2))
    ax.xaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle=':')
    ax.yaxis.grid(b=True, color=[0.3, 0.3, 0.3], linestyle='--')
    ax.set_ylim([0, 1.05 * (base + height)])
    if save_fig:
        if file_name is None:
            file_name = 'polar_recon_dirac.pdf'
        plt.savefig(file_name, format='pdf', dpi=300, transparent=True)


# ===========plotting functions===========
def sph_plot_diracs_plotly(
        colatitude_ref=None, azimuth_ref=None, colatitude=None, azimuth=None,
        dirty_img=None, azimuth_grid=None, colatitude_grid=None,
        surface_base=1, surface_height=0.,
        ):
    '''
    Plots a 2D map on a sphere as well as a collection of diracs using the plotly library

    Parameters
    ----------
    colatitude_ref: ndarray, optional
        The colatitudes of a collection of reference points
    azimuths_ref: ndarray, optional
        The azimuths of a collection of reference points for the Diracs
    colatitude: ndarray, optional
        The colatitudes of the collection of points to visualize
    azimuth: ndarray, optional
        The azimuths of the collection of points to visualize
    dirty_img: ndarray
        A 2D map for displaying a pattern on the sphere under the points
    azimuth_grid: ndarray
        The azimuths indexing the dirty_img 2D map
    colatitude_grid: ndarray
        The colatitudes indexing the dirty_img 2D map
    surface_base:
        radius corresponding to lowest height on the map
    sufrace_height:
        radius difference between the lowest and highest point on the map
    '''

    try:
        from plotly.offline import plot
        import plotly.graph_objs as go
        import plotly
    except ImportError:
        import warnings
        warnings.warn('The plotly package is required to use this function')
        return

    plotly.offline.init_notebook_mode()

    traces = []

    if dirty_img is not None and azimuth_grid is not None and colatitude_grid is not None:

        surfacecolor = np.abs(dirty_img)  # for plotting purposes

        base = surface_base

        surf_diff = surfacecolor.max() - surfacecolor.min()
        if surf_diff > 0:
            height = surface_height / surf_diff
        else:
            height = 0


        r_surf = base + height * (surfacecolor - surfacecolor.min()) / (surfacecolor.max() - surfacecolor.min())

        x_plt = r_surf * np.sin(colatitude_grid) * np.cos(azimuth_grid)
        y_plt = r_surf * np.sin(colatitude_grid) * np.sin(azimuth_grid)
        z_plt = r_surf * np.cos(colatitude_grid)

        trace1 = go.Surface(x=x_plt, y=y_plt, z=z_plt, surfacecolor=surfacecolor,
                            opacity=1, colorscale='Portland', hoverinfo='none')

        trace1['contours']['x']['highlightwidth'] = 1
        trace1['contours']['y']['highlightwidth'] = 1
        trace1['contours']['z']['highlightwidth'] = 1

        traces.append(trace1)

    if colatitude_ref is not None and azimuth_ref is not None:

        x_ref = np.sin(colatitude_ref) * np.cos(azimuth_ref)
        y_ref = np.sin(colatitude_ref) * np.sin(azimuth_ref)
        z_ref = np.cos(colatitude_ref)

        if not hasattr(colatitude_ref, '__iter__'):
            colatitude_ref = [colatitude_ref]
            azimuth_ref = [azimuth_ref]
            x_ref = [x_ref]
            y_ref = [y_ref]
            z_ref = [z_ref]

        text_str2 = []
        for count, colatitude0 in enumerate(colatitude_ref):
            text_str2.append(
                u'({0:.2f}\N{DEGREE SIGN}, {1:.2f}\N{DEGREE SIGN})'.format(np.degrees(colatitude0),
                                                                           np.degrees(azimuth_ref[count]))
            )

        trace2 = go.Scatter3d(mode='markers', name='ground truth',
                              x=x_ref, y=y_ref, z=z_ref,
                              text=text_str2,
                              hoverinfo='name+text',
                              marker=dict(size=6, symbol='circle', opacity=0.6,
                                          line=dict(
                                              color='rgb(204, 204, 204)',
                                              width=2
                                          ),
                                          color='rgb(0, 0.447, 0.741)'))
        traces.append(trace2)


    if colatitude is not None and azimuth is not None:

        x_recon = np.sin(colatitude) * np.cos(azimuth)
        y_recon = np.sin(colatitude) * np.sin(azimuth)
        z_recon = np.cos(colatitude)

        if not hasattr(colatitude, '__iter__'):
            colatitude_ref = [colatitude]
            azimuth = [azimuth]
            x_recon = [x_recon]
            y_recon = [y_recon]
            z_recon = [z_recon]

        text_str3 = []
        for count, colatitude0 in enumerate(colatitude):
            text_str3.append(
                u'({0:.2f}\N{DEGREE SIGN}, {1:.2f}\N{DEGREE SIGN})'.format(np.degrees(colatitude0),
                                                                           np.degrees(azimuth[count]))
            )

        trace3 = go.Scatter3d(mode='markers', name='reconstruction',
                              x=x_recon, y=y_recon, z=z_recon,
                              text=text_str3,
                              hoverinfo='name+text',
                              marker=dict(size=6, symbol='diamond', opacity=0.6,
                                          line=dict(
                                              color='rgb(204, 204, 204)',
                                              width=2
                                          ),
                                          color='rgb(0.850, 0.325, 0.098)'))
        traces.append(trace3)

    data = go.Data(traces)

    layout = go.Layout(title='', autosize=False,
                       width=670, height=550, showlegend=True,
                       margin=go.Margin(l=45, r=45, b=55, t=45)
                       )

    layout['legend']['xanchor'] = 'center'
    layout['legend']['yanchor'] = 'top'
    layout['legend']['x'] = 0.5

    fig = go.Figure(data=data, layout=layout)
    plot(fig)


def sph_plot_diracs(
        colatitude_ref=None, azimuth_ref=None, colatitude=None, azimuth=None,
        dirty_img=None, colatitude_grid=None, azimuth_grid=None,
            
                    file_name='sph_recon_2d_dirac.pdf', **kwargs):
    '''
    This function plots the dirty image with sources locations on
    a flat projection of the sphere

    Parameters
    ----------
    colatitude_ref: ndarray, optional
        The colatitudes of a collection of reference points
    azimuths_ref: ndarray, optional
        The azimuths of a collection of reference points for the Diracs
    colatitude: ndarray, optional
        The colatitudes of the collection of points to visualize
    azimuth: ndarray, optional
        The azimuths of the collection of points to visualize
    dirty_img: ndarray
        A 2D map for displaying a pattern on the sphere under the points
    azimuth_grid: ndarray
        The azimuths indexing the dirty_img 2D map
    colatitude_grid: ndarray
        The colatitudes indexing the dirty_img 2D map
    '''

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings
        warnings.warn('Matplotlib is required for plotting')
        return

    fig = plt.figure(figsize=(6.47, 4), dpi=90)
    ax = fig.add_subplot(111, projection="mollweide")

    if dirty_img is not None and colatitude_grid is not None and azimuth_grid is not None:

        azimuth_plt_internal = azimuth_grid.copy()
        azimuth_plt_internal[azimuth_grid > np.pi] -= np.pi * 2
        p_hd = ax.pcolormesh(azimuth_plt_internal, np.pi / 2. - colatitude_grid,
                             np.real(dirty_img), cmap='Spectral_r',
                             linewidth=0, alpha=0.3,  # shading='gouraud',
                             antialiased=True, zorder=0)

        p_hd.set_edgecolor('None')
        p_hdc = fig.colorbar(p_hd, orientation='horizontal', use_gridspec=False,
                             anchor=(0.5, 1.2), shrink=0.75, spacing='proportional')
        # p_hdc.formatter.set_powerlimits((0, 0))
        p_hdc.ax.tick_params(labelsize=8.5)
        p_hdc.update_ticks()

    if colatitude_ref is not None and azimuth_ref is not None:

        if hasattr(colatitude_ref, '__iter__'):
            K = colatitude_ref.size
            x_ref = azimuth_ref.copy()
            y_ref = np.pi * 0.5 - colatitude_ref.copy()  # convert CO-LATITUDE to LATITUDE

            ind = x_ref > np.pi
            x_ref[ind] -= 2 * np.pi
        else:
            K = 1
            x_ref = azimuth_ref
            y_ref = np.pi * 0.5 - colatitude_ref

            if x_ref > np.pi:
                x_ref = x_ref - 2 * np.pi

        ax.scatter(x_ref, y_ref,
                   c=np.tile([0, 0.447, 0.741], (K, 1)),
                   s=70, alpha=0.75, zorder=11,
                   marker='^', linewidths=0, cmap='Spectral_r',
                   label='Original')


    if colatitude is not None and azimuth is not None:

        if hasattr(colatitude, '__iter__'):
            K_est = colatitude.size
            x = azimuth.copy()
            y = np.pi * 0.5 - colatitude.copy()  # convert CO-LATITUDE to LATITUDE

            ind = x > np.pi
            x[ind] -= 2 * np.pi  # scale conversion to -pi to pi
        else:
            K_est = 1
            x = azimuth
            y = np.pi * 0.5 - colatitude

            if x > np.pi:
                x = x - 2 * np.pi

        ax.scatter(x, y,
                   c=np.tile([0.850, 0.325, 0.098], (K_est, 1)),
                   s=100, alpha=0.75, zorder=12,
                   marker='*', linewidths=0, cmap='Spectral_r',
                   label='Reconstruction')

    ax.set_xticklabels([u'210\N{DEGREE SIGN}', u'240\N{DEGREE SIGN}',
                        u'270\N{DEGREE SIGN}', u'300\N{DEGREE SIGN}',
                        u'330\N{DEGREE SIGN}', u'0\N{DEGREE SIGN}',
                        u'30\N{DEGREE SIGN}', u'60\N{DEGREE SIGN}',
                        u'90\N{DEGREE SIGN}', u'120\N{DEGREE SIGN}',
                        u'150\N{DEGREE SIGN}'])

    ax.legend(scatterpoints=1, loc=8, fontsize=9,
              ncol=2, bbox_to_anchor=(0.5, -0.18),
              handletextpad=.2, columnspacing=1.7, labelspacing=0.1)  # framealpha=0.3,

    if 'title_str' in kwargs:
        ax.set_title(title_str, fontsize=11)

    ax.set_xlabel(r'azimuth', fontsize=11)
    ax.set_ylabel(r'latitude', fontsize=11)

    ax.xaxis.set_label_coords(0.5, 0.52)

    ax.grid(True)

    '''
    if save_fig:
        plt.savefig(file_name, format='pdf', dpi=300, transparent=True)
    '''




'''
Adaptive Filter Algorithms
==========================

This sub-package provides implementations of popular adaptive filter algorithms.

RLS
    | Recursive Least Squares
    | :py:obj:`pyroomacoustics.adaptive.rls` 
LMS
    | Least Mean Squares and Normalized Least Mean Squares
    | :py:obj:`pyroomacoustics.adaptive.lms`

All these classes derive from the base class
:py:obj:`pyroomacoustics.adaptive.adaptive_filter.AdaptiveFilter` that offer
a generic way of running an adaptive filter.

The above classes are applicable for time domain processing. For frequency 
domain adaptive filtering, there is the SubbandLMS class. After using a DFT or
STFT block, the SubbandLMS class can be used to used to apply LMS or NLMS to 
each frequency band. A shorter adaptive filter can be used on each band as 
opposed to the filter required in the time domain version. Roughly, a filter of
M taps applied to each band (total of B) corresponds to a time domain filter 
with N = M x B taps.

How to use the adaptive filter module
-------------------------------------

First, an adaptive filter object is created and all the relevant options
can be set (step size, regularization, etc). Then, the update function
is repeatedly called to provide new samples to the algorithm.

::

    # initialize the filter
    rls = pyroomacoustics.adaptive.RLS(30)

    # run the filter on a stream of samples
    for i in range(100):
        rls.update(x[i], d[i])

    # the reconstructed filter is available
    print('Reconstructed filter:', rls.w)


The SubbandLMS class has the same methods as the time domain 
approaches. However, the signal must be in the frequency domain. This
can be done with the STFT block in the `transform` sub-package of
`pyroomacoustics`.

::

    # initialize STFT and SubbandLMS blocks
    block_size = 128
    stft_x = pra.transform.STFT(N=block_size,
        hop=block_size//2, 
        analysis_window=pra.hann(block_size))
    stft_d = pra.transform.STFT(N=block_size,
        hop=block_size//2, 
        analysis_window=pra.hann(block_size))
    nlms = pra.adaptive.SubbandLMS(num_taps=6, 
        num_bands=block_size//2+1, mu=0.5, nlms=True)

    # preparing input and reference signals
    ...
    
    # apply block-by-block
    for n in range(num_blocks):

        # obtain block
        ...

        # to frequency domain
        stft_x.analysis(x_block)
        stft_d.analysis(d_block)
        nlms.update(stft_x.X, stft_d.X)

        # estimating input convolved with unknown response
        y_hat = stft_d.synthesis(np.diag(np.dot(nlms.W.conj().T,stft_x.X)))

        # AEC output
        E = stft_d.X - np.diag(np.dot(nlms.W.conj().T,stft_x.X))
        out = stft_d.synthesis(E)


Other Available Subpackages
---------------------------

:py:obj:`pyroomacoustics.adaptive.data_structures`
    this provides abstractions for computing functions on regular or irregular
    grids defined on circles and spheres with peak finding methods

:py:obj:`pyroomacoustics.adaptive.util`
    a few methods mainly to efficiently manipulate Toeplitz and Hankel matrices

Utilities
---------

:py:obj:`pyroomacoustics.adaptive.algorithms`
    a dictionary containing all the adaptive filter object subclasses availables indexed by
    keys ``['RLS', 'BlockRLS', 'BlockLMS', 'NLMS', 'SubbandLMS']``

'''

from .adaptive_filter import *
from .lms import *
from .rls import *
from .subband_lms import *
from .util import *
from .data_structures import *

# Create this dictionary as a shortcut to different algorithms
algorithms = {
        'RLS' : RLS,
        'BlockRLS' : BlockRLS,
        'NLMS' : NLMS,
        'BlockLMS' : BlockLMS,
        'SubbandLMS' : SubbandLMS
        }




# Author: Eric Bezzam
# Date: July 15, 2016

import numpy as np

from .doa import DOA


class MUSIC(DOA):
    """
    Class to apply MUltiple SIgnal Classication (MUSIC) direction-of-arrival
    (DoA) for a particular microphone array.

    .. note:: Run locate_source() to apply the MUSIC algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    """

    def __init__(
        self,
        L,
        fs,
        nfft,
        c=343.0,
        num_src=1,
        mode="far",
        r=None,
        azimuth=None,
        colatitude=None,
        **kwargs
    ):

        DOA.__init__(
            self,
            L=L,
            fs=fs,
            nfft=nfft,
            c=c,
            num_src=num_src,
            mode=mode,
            r=r,
            azimuth=azimuth,
            colatitude=colatitude,
            **kwargs
        )

        self.Pssl = None

    def _process(self, X):
        """
        Perform MUSIC for given frame in order to estimate steered response
        spectrum.
        """
        # compute steered response
        self.Pssl = np.zeros((self.num_freq, self.grid.n_points))
        C_hat = self._compute_correlation_matricesvec(X)
        # subspace decomposition
        Es, En, ws, wn = self._subspace_decomposition(C_hat[None, ...])
        # compute spatial spectrum
        identity = np.zeros((self.num_freq, self.M, self.M))
        identity[:, list(np.arange(self.M)), list(np.arange(self.M))] = 1
        cross = identity - np.matmul(Es, np.moveaxis(np.conjugate(Es), -1, -2))
        self.Pssl = self._compute_spatial_spectrumvec(cross)
        self.grid.set_values(np.squeeze(np.sum(self.Pssl, axis=1) / self.num_freq))

    def plot_individual_spectrum(self):
        """
        Plot the steered response for each frequency.
        """

        # check if matplotlib imported
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        # only for 2D
        if self.grid.dim == 3:
            pass
        else:
            import warnings

            warnings.warn("Only for 2D.")
            return

        # plot
        for k in range(self.num_freq):

            freq = float(self.freq_bins[k]) / self.nfft * self.fs
            azimuth = self.grid.azimuth * 180 / np.pi

            plt.plot(azimuth, self.Pssl[k, 0 : len(azimuth)])

            plt.ylabel("Magnitude")
            plt.xlabel("Azimuth [degrees]")
            plt.xlim(min(azimuth), max(azimuth))
            plt.title("Steering Response Spectrum - " + str(freq) + " Hz")
            plt.grid(True)

    def _compute_spatial_spectrumvec(self, cross):
        mod_vec = np.transpose(
            np.array(self.mode_vec[self.freq_bins, :, :]), axes=[2, 0, 1]
        )
        # timeframe, frequ, no idea
        denom = np.matmul(
            np.conjugate(mod_vec[..., None, :]), np.matmul(cross, mod_vec[..., None])
        )
        return 1.0 / abs(denom[..., 0, 0])

    def _compute_spatial_spectrum(self, cross, k):

        P = np.zeros(self.grid.n_points)

        for n in range(self.grid.n_points):
            Dc = np.array(self.mode_vec[k, :, n], ndmin=2).T
            Dc_H = np.conjugate(np.array(self.mode_vec[k, :, n], ndmin=2))
            denom = np.dot(np.dot(Dc_H, cross), Dc)
            P[n] = 1 / abs(denom)

        return P

    # non-vectorized version
    def _compute_correlation_matrices(self, X):
        C_hat = np.zeros([self.num_freq, self.M, self.M], dtype=complex)
        for i in range(self.num_freq):
            k = self.freq_bins[i]
            for s in range(self.num_snap):
                C_hat[i, :, :] = C_hat[i, :, :] + np.outer(
                    X[:, k, s], np.conjugate(X[:, k, s])
                )
        return C_hat / self.num_snap

    # vectorized version
    def _compute_correlation_matricesvec(self, X):
        # change X such that time frames, frequency microphones is the result
        X = np.transpose(X, axes=[2, 1, 0])
        # select frequency bins
        X = X[..., list(self.freq_bins), :]
        # Compute PSD and average over time frame
        C_hat = np.matmul(X[..., None], np.conjugate(X[..., None, :]))
        # Average over time-frames
        C_hat = np.mean(C_hat, axis=0)
        return C_hat

    # vectorized versino
    def _subspace_decomposition(self, R):
        # eigenvalue decomposition!
        # This method is specialized for Hermitian symmetric matrices,
        # which is the case since R is a covariance matrix
        w, v = np.linalg.eigh(R)

        # This method (numpy.linalg.eigh) returns the eigenvalues (and
        # eigenvectors) in ascending order, so there is no need to sort Signal
        # comprises the leading eigenvalues Noise takes the rest

        Es = v[..., -self.num_src :]
        ws = w[..., -self.num_src :]
        En = v[..., : -self.num_src]
        wn = w[..., : -self.num_src]

        return (Es, En, ws, wn)




import numpy as np
import pyroomacoustics as pra
from scipy.io import wavfile

# We use several sound samples for each source to have a long enough length
wav_files = [
        [
            'examples/input_samples/cmu_arctic_us_axb_a0004.wav',
            'examples/input_samples/cmu_arctic_us_axb_a0005.wav',
            'examples/input_samples/cmu_arctic_us_axb_a0006.wav',
            ],
        [
            'examples/input_samples/cmu_arctic_us_aew_a0001.wav',
            'examples/input_samples/cmu_arctic_us_aew_a0002.wav',
            'examples/input_samples/cmu_arctic_us_aew_a0003.wav',
            ],
        ]

def test_trinicon():
    '''
    This test do not check the correctness of the output of Trinicon.
    Only that it runs without errors.
    '''

    # STFT frame length
    L = 1024

    # Room 8m by 9m
    room_dim = [8, 9]

    # source location
    source = np.array([1, 4.5])

    # create the room with sources and mics
    room = pra.ShoeBox(
        room_dim,
        fs=16000,
        max_order=0,
        sigma2_awgn=1e-8)

    # get signals
    signals = [ np.concatenate([wavfile.read(f)[1].astype(np.float32)
        for f in source_files])
        for source_files in wav_files ]
    delays = [1., 0.]
    locations = [[2.5,3], [2.5, 6]]

    # add mic and good source to room
    # Add silent signals to all sources
    for sig, d, loc in zip(signals, delays, locations):
        room.add_source(loc, signal=np.zeros_like(sig), delay=d)

    # add microphone array
    room.add_microphone_array(
            pra.MicrophoneArray(np.c_[[6.5, 4.39], [6.5, 4.51]], fs=room.fs)
            )

    # compute RIRs
    room.compute_rir()

    # Record each source separately
    separate_recordings = []
    for source, signal in zip(room.sources, signals):

        source.signal[:] = signal

        room.simulate()
        separate_recordings.append(room.mic_array.signals)

        source.signal[:] = 0.
    separate_recordings = np.array(separate_recordings)

    # Mix down the recorded signals
    mics_signals = np.sum(separate_recordings, axis=0)

    # run Trinicon
    y,w = pra.bss.trinicon(mics_signals, filter_length=L, return_filters=True)


if __name__ == '__main__':
    # STFT frame length
    L = 1024

    # Room 8m by 9m
    room_dim = [8, 9]

    # source location
    source = np.array([1, 4.5])

    # create the room with sources and mics
    room = pra.ShoeBox(
        room_dim,
        fs=16000,
        max_order=0,
        sigma2_awgn=1e-8)

    # get signals
    signals = [ np.concatenate([wavfile.read(f)[1].astype(np.float32)
        for f in source_files])
        for source_files in wav_files ]
    delays = [1., 0.]
    locations = [[2.5,3], [2.5, 6]]

    # add mic and good source to room
    # Add silent signals to all sources
    for sig, d, loc in zip(signals, delays, locations):
        room.add_source(loc, signal=np.zeros_like(sig), delay=d)

    # add microphone array
    room.add_microphone_array(
            pra.MicrophoneArray(np.c_[[6.5, 4.39], [6.5, 4.51]], fs=room.fs)
            )

    # compute RIRs
    room.compute_rir()

    # Record each source separately
    separate_recordings = []
    for source, signal in zip(room.sources, signals):

        source.signal[:] = signal

        room.simulate()
        separate_recordings.append(room.mic_array.signals)

        source.signal[:] = 0.
    separate_recordings = np.array(separate_recordings)

    # Mix down the recorded signals
    mics_signals = np.sum(separate_recordings, axis=0)

    # START BSS
    ###########

    y,w = pra.bss.trinicon(mics_signals, filter_length=L, return_filters=True)
    y,w = pra.bss.trinicon(mics_signals, w0=w, return_filters=True)
    y,w = pra.bss.trinicon(mics_signals, w0=w, return_filters=True)
    y,w = pra.bss.trinicon(mics_signals, w0=w, return_filters=True)
    y,w = pra.bss.trinicon(mics_signals, w0=w, return_filters=True)
    y,w = pra.bss.trinicon(mics_signals, w0=w, return_filters=True)
    y = pra.bss.trinicon(mics_signals, w0=w)

    # Compare SIR
    #############
    sdr, isr, sir, sar, perm = bss_eval_images(ref[:,:,0], y[:,L//2:ref.shape[1]+L//2])

    print('SDR:', sdr)
    print('SIR:', sir)

    import matplotlib.pyplot as plt
    plt.figure()
    plt.subplot(2,2,1)
    plt.specgram(ref[0,:,0], NFFT=1024, Fs=room.fs)
    plt.title('Reference 1')

    plt.subplot(2,2,2)
    plt.specgram(ref[1,:,0], NFFT=1024, Fs=room.fs)
    plt.title('Reference 2')

    plt.subplot(2,2,3)
    plt.specgram(y[perm[0],:], NFFT=1024, Fs=room.fs)
    plt.title('Source 1')

    plt.subplot(2,2,4)
    plt.specgram(y[perm[1],:], NFFT=1024, Fs=room.fs)
    plt.title('Source 2')

    plt.figure()
    a = np.array(SDR)
    b = np.array(SIR)
    plt.plot(np.arange(a.shape[0]) * 10, a, label='SDR')
    plt.plot(np.arange(b.shape[0]) * 10, b, label='SIR')
    plt.legend()

    plt.show()

    import sounddevice as sd
    def play(ch):
        sd.play(pra.normalize(y[ch]) * 0.75, samplerate=room.fs, blocking=True)





# Author: Eric Bezzam
# Last modification: February 22, 2019

"""
Class for performing the Discrete Fourier Transform (DFT) and inverse DFT for
real signals, including multichannel. It is also possible to specific an 
analysis or synthesis window.

When available, it is possible to use the ``pyfftw`` or ``mkl_fft`` packages.
Otherwise the default is to use ``numpy.fft.rfft``/``numpy.fft.irfft``.

More on ``pyfftw`` can be read here: 
https://pyfftw.readthedocs.io/en/latest/index.html

More on ``mkl_fft`` can be read here: 
https://github.com/IntelPython/mkl_fft
"""

import numpy as np
from numpy.fft import rfft, irfft
import warnings

try:
    import pyfftw
    pyfftw_available = True
except ImportError:
    pyfftw_available = False

try:
    import mkl_fft    # https://github.com/IntelPython/mkl_fft
    mkl_available = True
except ImportError:
    mkl_available = False


class DFT(object):
    """
    Class for performing the Discrete Fourier Transform (DFT) of real signals.

    Attributes
    ----------
    X: numpy array
        Real DFT computed by ``analysis``.
    x: numpy array
        IDFT computed by ``synthesis``.
    nfft: int
        FFT size.
    D: int
        Number of channels.
    transform: str
        Which FFT package will be used.
    analysis_window: numpy array
        Window to be applied before DFT.
    synthesis_window: numpy array
        Window to be applied after inverse DFT.
    axis : int
        Axis over which to compute the FFT.
    precision, bits : string, np.float32, np.float64, np.complex64, np.complex128
        How many precision bits to use for the input. Twice the amount will be used
        for complex spectrum.

    Parameters
    ----------
    nfft: int
        FFT size.
    D: int, optional
        Number of channels. Default is 1.
    analysis_window: numpy array, optional
        Window to be applied before DFT. Default is no window.
    synthesis_window: numpy array, optional
        Window to be applied after inverse DFT. Default is no window.
    transform: str, optional
        which FFT package to use: ``numpy``, ``pyfftw``, or ``mkl``. Default is 
        ``numpy``.
    axis : int, optional
        Axis over which to compute the FFT. Default is first axis.
    precision : string, np.float32, np.float64, np.complex64, np.complex128, optional
        How many precision bits to use for the input.
        If 'single'/np.float32/np.complex64, 32 bits for real inputs or 64 for complex spectrum.
        Otherwise, cast to 64 bits for real inputs or 128 for complex spectrum (default).

    """

    def __init__(self, nfft, D=1, analysis_window=None, synthesis_window=None, 
        transform='numpy', axis=0, precision='double', bits=None):

        self.nfft = nfft
        self.D = D
        self.axis=axis

        if bits is not None and precision is not None:
            warnings.warn('Deprecated keyword "bits" ignored in favor of new keyword "precision"', DeprecationWarning)
        elif bits is not None and precision is None:
            warnings.warn('Keyword "bits" is deprecated and has been replaced by "precision"')
            if bits == 32:
                precision = 'single'
            elif bits == 64:
                precision = 'double'

        if precision == np.float32 or precision == np.complex64 or precision == 'single':
            time_dtype = np.float32
            freq_dtype = np.complex64
        else:
            time_dtype = np.float64
            freq_dtype = np.complex128

        if axis==0:
            self.x = np.squeeze(np.zeros((self.nfft, self.D), 
                dtype=time_dtype))
            self.X = np.squeeze(np.zeros((self.nfft//2+1, self.D),
                dtype=freq_dtype))
        elif axis==1:
            self.x = np.squeeze(np.zeros((self.D, self.nfft),
                dtype=time_dtype))
            self.X = np.squeeze(np.zeros((self.D, self.nfft//2+1),
                dtype=freq_dtype))
        else:
            raise ValueError("Invalid 'axis' option. Must be 0 or 1.")

        if analysis_window is not None:
            if axis==0 and D>1:
                self.analysis_window = analysis_window[:,np.newaxis].astype(time_dtype)
            else:
                self.analysis_window = analysis_window.astype(time_dtype)
        else:
            self.analysis_window = None

        if synthesis_window is not None:
            if axis==0 and D>1:
                self.synthesis_window = synthesis_window[:,np.newaxis].astype(time_dtype)
            else:
                self.synthesis_window = synthesis_window.astype(time_dtype)
        else:
            self.synthesis_window=None


        if transform == 'fftw':
            if pyfftw_available:
                from pyfftw import empty_aligned, FFTW
                self.transform = transform
                # allocate input (real) and output for pyfftw
                if self.D==1:
                    self.a = empty_aligned(self.nfft, dtype=time_dtype)
                    self.b = empty_aligned(self.nfft//2+1, dtype=freq_dtype)
                    self._forward = FFTW(self.a, self.b)
                    self._backward = FFTW(self.b, self.a, 
                        direction='FFTW_BACKWARD')
                else:
                    if axis==0:
                        self.a = empty_aligned([self.nfft, self.D], 
                            dtype=time_dtype)
                        self.b = empty_aligned([self.nfft//2+1, self.D], 
                            dtype=freq_dtype)
                    elif axis==1:
                        self.a = empty_aligned([self.D, self.nfft], 
                            dtype=time_dtype)
                        self.b = empty_aligned([self.D, self.nfft//2+1], 
                            dtype=freq_dtype)
                    self._forward = FFTW(self.a, self.b, 
                        axes=(self.axis, ))
                    self._backward = FFTW(self.b, self.a, axes=(self.axis, ), 
                        direction='FFTW_BACKWARD')
            else: 
                warnings.warn("Could not import pyfftw wrapper for fftw functions. Using numpy's rfft instead.")
                self.transform = 'numpy'
        elif transform == 'mkl':
            if mkl_available:
                import mkl_fft
                self.transform = 'mkl'
            else:
                warnings.warn("Could not import mkl wrapper. Using numpy's rfft instead.")
                self.transform = 'numpy'
        else:
            self.transform = 'numpy'


    def analysis(self, x):
        """
        Perform frequency analysis of a real input using DFT.

        Parameters
        ----------
        x : numpy array
            Real signal in time domain.

        Returns
        -------
        numpy array
            DFT of input.
        """

        # check for valid input
        if x.shape != self.x.shape:
            raise ValueError('Invalid input dimensions! Got (%d, %d), expecting (%d, %d).' 
                % (x.shape[0], x.shape[1],
                self.x.shape[0], self.x.shape[1]))

        # apply window if needed
        if self.analysis_window is not None:
            np.multiply(self.analysis_window, x, x)

        # apply DFT
        if self.transform == 'fftw':
            self.a[:,] = x
            self.X[:,] = self._forward()
        elif self.transform == 'mkl':
            self.X[:,] = mkl_fft.rfft_numpy(x, self.nfft, axis=self.axis)
        else:
            self.X[:,] = rfft(x, self.nfft, axis=self.axis)

        return self.X


    def synthesis(self, X=None):
        """
        Perform time synthesis of frequency domain to real signal using the 
        inverse DFT.

        Parameters
        ----------
        X : numpy array, optional
            Complex signal in frequency domain. Default is to use DFT computed
            from ``analysis``.

        Returns
        -------
        numpy array
            IDFT of ``self.X`` or input if given.
        """

        # check for valid input
        if X is not None:
            if X.shape != self.X.shape:
                raise ValueError('Invalid input dimensions! Got (%d, %d), expecting (%d, %d).' 
                    % (X.shape[0], X.shape[1],
                    self.X.shape[0], self.X.shape[1]))

            self.X[:,] = X

        # inverse DFT
        if self.transform == 'fftw':
            self.b[:] = self.X
            self.x[:,] = self._backward()
        elif self.transform == 'mkl':
            self.x[:,] = mkl_fft.irfft_numpy(self.X, self.nfft, axis=self.axis)
        else:
            self.x[:,] = irfft(self.X, self.nfft, axis=self.axis)

        # apply window if needed
        if self.synthesis_window is not None:
            np.multiply(self.synthesis_window, self.x, self.x)

        return self.x



# coding=utf-8
#
# MIT License
#
# Window functions Copyright (C) 2015-2019 Taishi Nakashima, Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

r"""
Window Functions
================

This is a collection of many popular window functions used in signal processing.

A few options are provided to correctly construct the required window function.
The ``flag`` keyword argument can take the following values.

``asymmetric``
  This way, many of the functions will sum to one when their left part is added
  to their right part. This is useful for overlapped transforms such as the STFT.

``symmetric``
  With this flag, the window is perfectly symmetric. This might be more
  suitable for analysis tasks.
``mdct``
  Available for only some of the windows. The window is modified to satisfy
  the perfect reconstruction condition of the MDCT transform.

Often, we would like to get the full window function, but on some occasions, it is useful
to get only the left (or right) part. This can be indicated via the keyword argument
``length`` that can take values ``full`` (default), ``left``, or ``right``.
"""

import numpy as np
from scipy import special
pi = np.pi


# Bartlett window
def bart(N, flag='asymmetric', length='full'):
    r'''
    The Bartlett window function

    .. math::

        w[n] = 2 / (M-1) ((M-1)/2 - |n - (M-1)/2|) , n=0,\ldots,N-1


    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 2/(N-1) * ((N-1)/2 - np.abs(t - (N-1)/2))

    # make the window respect MDCT condition
    if flag == 'mdct':
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Modified Bartlett--Hann window
def bart_hann(N, flag='asymmetric', length='full'):
    r'''
    The modified Bartlett--Hann window function

    .. math::

        w[n] = 0.62 - 0.48|(n/M-0.5)| + 0.38 \cos(2\pi(n/M-0.5)),
        n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.62 - 0.48 * np.abs(t/N - 0.5) + 0.38 * np.cos(2*pi*(t/N - 0.5))

    # make the window respect MDCT condition
    if flag == 'mdct':
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Blackman window
def blackman(N, flag='asymmetric', length='full'):
    r'''
    The Blackman window function

    .. math::

        w[n] = 0.42 - 0.5\cos(2\pi n/(M-1)) + 0.08\cos(4\pi n/(M-1)),
        n = 0, \ldots, M-1


    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.42 - 0.5*np.cos(2*pi*t/(N-1)) + 0.08*np.cos(4*pi*t/(N-1))

    # make the window respect MDCT condition
    if flag == 'mdct':
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Blackman-Harris window
def blackman_harris(N, flag='asymmetric', length='full'):
    r'''
    The Hann window function

    .. math::

        w[n] = a_0 - a_1 \cos(2\pi n/M)
        + a_2 \cos(4\pi n/M) + a_3 \cos(6\pi n/M), n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # coefficients
    a = np.array([.35875, .48829, .14128, .01168])

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag == 'symmetric':
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = a[0] - a[1]*np.cos(2*pi*t) + a[2]*np.cos(4*pi*t) + a[3]*np.cos(6*pi*t)

    return w


# Bohman window function
def bohman(N, flag='asymmetric', length='full'):
    r'''
    The Bohman window function

    .. math::

        w[n] = (1-|x|) \cos(\pi |x|) + \pi / |x| \sin(\pi |x|), -1\leq x\leq 1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    x = np.abs(np.linspace(-1, 1, N)[1:-1])
    w = (1 - x) * np.cos(pi * x) + 1.0 / pi * np.sin(pi * x)
    w = np.r_[0, w, 0]

    # make the window respect MDCT condition
    if flag == 'mdct':
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# cosine window function
def cosine(N, flag='asymmetric', length='full'):
    r'''
    The cosine window function

    .. math::

        w[n] = \cos(\pi (n/M - 0.5))^2

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = np.cos(pi * (t - 0.5)) ** 2

    # make the window respect MDCT condition
    if flag == 'mdct':
        w **= 2
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Flattop window
def flattop(N, flag='asymmetric', length='full'):
    r'''
    The flat top weighted window function

    .. math::

        w[n] = a_0 - a_1 \cos(2\pi n/M) + a_2 \cos(4\pi n/M)
        + a_3 \cos(6\pi n/M) + a_4 \cos(8\pi n/M), n=0,\ldots,N-1

    where

    .. math::
        a0 = 0.21557895
        a1 = 0.41663158
        a2 = 0.277263158
        a3 = 0.083578947
        a4 = 0.006947368

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are
          used for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # coefficients
    a = np.array([.21557895, .41663158, .277263158, .083578947, .006947368])

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag == 'symmetric':
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = a[0] - a[1]*np.cos(2*pi*t) + a[2]*np.cos(4*pi*t)\
        + a[3]*np.cos(6*pi*t) + a[4]*np.cos(8*pi*t)

    return w


# Gaussian window
def gaussian(N, std, flag='asymmetric', length='full'):
    r'''
    The flat top weighted window function

    .. math::
        w[n] = e^{ -\frac{1}{2}\left(\frac{n}{\sigma}\right)^2 }

    Parameters
    ----------
    N: int
        the window length
    std: float
        the standard deviation
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag == 'symmetric':
        t = t / float(N - 1)
    else:
        t = t / float(N)

    n = np.arange(0, N) - (N - 1.0) / 2.0
    sig2 = 2 * std**2
    w = np.exp(-n**2 / sig2)

    return w


# hamming window function
def hamming(N, flag='asymmetric', length='full'):
    r'''
    The Hamming window function

    .. math::

        w[n] = 0.54  - 0.46 \cos(2 \pi n / M), n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.54 - 0.46*np.cos(2*pi*t)

    # make the window respect MDCT condition
    if flag == 'mdct':
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# hann window function
def hann(N, flag='asymmetric', length='full'):
    r'''
    The Hann window function

    .. math::

        w[n] = 0.5 (1 - \cos(2 \pi n / M)), n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 0.5 * (1 - np.cos(2 * pi * t))

    # make the window respect MDCT condition
    if flag == 'mdct':
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Kaiser window function
def kaiser(N, beta, flag='asymmetric', length='full'):
    r'''
    The Kaiser window function

    .. math::
       w[n] = I_0\left( \beta \sqrt{1-\frac{4n^2}{(M-1)^2}} \right)/I_0(\beta)

    with

    .. math::
       \quad -\frac{M-1}{2} \leq n \leq \frac{M-1}{2},

    where :math:`I_0` is the modified zeroth-order Bessel function.

    Parameters
    ----------
    N: int
        the window length
    beta: float
        Shape parameter, determines trade-off between main-lobe width and
        side lobe level. As beta gets large, the window narrows.
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    n = np.arange(0, N)
    alpha = (N - 1) / 2.0
    w = (special.i0(beta * np.sqrt(1 - ((n - alpha) / alpha) ** 2.0)) /
         special.i0(beta))

    # make the window respect MDCT condition
    if flag == 'mdct':
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w


# Rectangular window function
def rect(N):
    r'''
    The rectangular window

    .. math::

        w[n] = 1, n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    '''

    return np.ones(N)


# triangular window function
def triang(N, flag='asymmetric', length='full'):
    r'''
    The triangular window function

    .. math::

        w[n] = 1 - | 2 n / M - 1 |, n=0,\ldots,N-1

    Parameters
    ----------
    N: int
        the window length
    flag: string, optional
        Possible values

        - *asymmetric*: asymmetric windows are used
          for overlapping transforms (:math:`M=N`)
        - *symmetric*: the window is symmetric (:math:`M=N-1`)
        - *mdct*: impose MDCT condition on the window (:math:`M=N-1` and
          :math:`w[n]^2 + w[n+N/2]^2=1`)

    length: string, optional
        Possible values

        - *full*: the full length window is computed
        - *right*: the right half of the window is computed
        - *left*: the left half of the window is computed
    '''

    # first choose the indexes of points to compute
    if length == 'left':     # left side of window
        t = np.arange(0, N / 2)
    elif length == 'right':   # right side of window
        t = np.arange(N / 2, N)
    else:                   # full window by default
        t = np.arange(0, N)

    # if asymmetric window, denominator is N, if symmetric it is N-1
    if flag in ['symmetric', 'mdct']:
        t = t / float(N - 1)
    else:
        t = t / float(N)

    w = 1. - np.abs(2. * t - 1.)

    # make the window respect MDCT condition
    if flag == 'mdct':
        d = w[:N / 2] + w[N / 2:]
        w[:N / 2] *= 1. / d
        w[N / 2:] *= 1. / d

    # compute window
    return w



# Author: Eric Bezzam
# Date: July 15, 2016
from __future__ import division, print_function

from .doa import *

class SRP(DOA):
    """
    Class to apply Steered Response Power (SRP) direction-of-arrival (DoA) for 
    a particular microphone array.

    .. note:: Run locate_source() to apply the SRP-PHAT algorithm.

    Parameters
    ----------
    L: numpy array
        Microphone array positions. Each column should correspond to the 
        cartesian coordinates of a single microphone.
    fs: float
        Sampling frequency.
    nfft: int
        FFT length.
    c: float
        Speed of sound. Default: 343 m/s
    num_src: int
        Number of sources to detect. Default: 1
    mode: str
        'far' or 'near' for far-field or near-field detection 
        respectively. Default: 'far'
    r: numpy array
        Candidate distances from the origin. Default: np.ones(1)
    azimuth: numpy array
        Candidate azimuth angles (in radians) with respect to x-axis.
        Default: np.linspace(-180.,180.,30)*np.pi/180
    colatitude: numpy array
        Candidate elevation angles (in radians) with respect to z-axis.
        Default is x-y plane search: np.pi/2*np.ones(1)
    """
    def __init__(self, L, fs, nfft, c=343.0, num_src=1, mode='far', r=None, 
        azimuth=None, colatitude=None, **kwargs):

        DOA.__init__(self, L=L, fs=fs, nfft=nfft, c=c, num_src=num_src, 
            mode=mode, r=r, azimuth=azimuth, colatitude=colatitude, **kwargs)

        self.num_pairs = self.M*(self.M-1)/2

        #self.mode_vec = np.conjugate(self.mode_vec)

    def _process(self, X):
        """
        Perform SRP-PHAT for given frame in order to estimate steered response 
        spectrum.
        """

        ones = np.ones(self.L.shape[1])

        srp_cost = np.zeros(self.grid.n_points)

        # apply PHAT weighting
        absX = np.abs(X)
        absX[absX < tol] = tol
        pX = X / absX

        CC = []
        for k in self.freq_bins:
            CC.append( np.dot(pX[:,k,:], np.conj(pX[:,k,:]).T) )
        CC = np.array(CC)

        for n in range(self.grid.n_points):

            # get the mode vector axis: (frequency, microphones)
            mode_vec = self.mode_vec[self.freq_bins,:,n]

            # compute the outer product along the microphone axis
            mode_mat = np.conj(mode_vec[:,:,None]) * mode_vec[:,None,:]

            # multiply covariance by mode vectors and sum over the frequencies
            R = np.sum(CC * mode_mat, axis=0)

            # Now sum over all distince microphone pairs
            sum_val = np.inner(ones, np.dot(np.triu(R, 1), ones))

            # Finally normalize
            srp_cost[n] = np.abs(sum_val) / self.num_snap/self.num_freq/self.num_pairs

        self.grid.set_values(srp_cost)



'''
pyroomacoustics
===============

Provides
  1. Room impulse simulations via the image source model
  2. Simulation of sound propagation using STFT engine
  3. Reference implementations of popular algorithms for

    * beamforming
    * direction of arrival
    * adaptive filtering
    * source separation
    * single channel denoising
    * etc

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided
with the code, and a loose standing reference guide, available from
`the pyroomacoustics readthedocs page <http://pyroomacoustics.readthedocs.io>`_.

We recommend exploring the docstrings using
`IPython <http://ipython.scipy.org>`_, an advanced Python shell with
TAB-completion and introspection capabilities.  See below for further
instructions.

The docstring examples assume that `pyroomacoustics` has been imported as `pra`::

  >>> import pyroomacoustics as pra

Code snippets are indicated by three greater-than signs::

  >>> x = 42
  >>> x = x + 1

Use the built-in ``help`` function to view a function's docstring::

  >>> help(pra.stft.STFT)
  ... # doctest: +SKIP

Available submodules
---------------------
:py:obj:`pyroomacoustics.acoustics`
    Acoustics and psychoacoustics routines, mel-scale, critcal bands, etc.

:py:obj:`pyroomacoustics.beamforming`
    Microphone arrays and beamforming routines.

:py:obj:`pyroomacoustics.geometry`
    Core geometry routine for the image source model.

:py:obj:`pyroomacoustics.metrics`
    Performance metrics like mean-squared error, median, Itakura-Saito, etc.

:py:obj:`pyroomacoustics.multirate`
    Rate conversion routines.

:py:obj:`pyroomacoustics.parameters`
    Global parameters, i.e. for physics constants.

:py:obj:`pyroomacoustics.recognition`
    Hidden Markov Model and TIMIT database structure.

:py:obj:`pyroomacoustics.room`
    Abstraction of room and image source model.

:py:obj:`pyroomacoustics.soundsource`
    Abstraction for a sound source.

:py:obj:`pyroomacoustics.stft`
    **Deprecated** Replaced by the methods in :py:obj:`pyroomacoustics.transform`

:py:obj:`pyroomacoustics.sync`
    A few routines to help synchronize signals.

:py:obj:`pyroomacoustics.utilities`
    A bunch of routines to do various stuff.

:py:obj:`pyroomacoustics.wall`
    Abstraction for walls of a room.

:py:obj:`pyroomacoustics.windows`
    Tapering windows for spectral analysis.

Available subpackages
---------------------

:py:obj:`pyroomacoustics.adaptive`
    Adaptive filter algorithms

:py:obj:`pyroomacoustics.bss`
    Blind source separation.

:py:obj:`pyroomacoustics.datasets`
    Wrappers around a few popular speech datasets

:py:obj:`pyroomacoustics.denoise`
    Single channel noise reduction methods

:py:obj:`pyroomacoustics.doa`
    Direction of arrival finding algorithms

:py:obj:`pyroomacoustics.phase`
    Phase-related processing

:py:obj:`pyroomacoustics.transform`
    Block frequency domain processing tools


Utilities
---------
__version__
    pyroomacoustics version string

'''

from .version import __version__

from . import libroom as libroom

from .room import *
from .beamforming import *
from .soundsource import *
from .parameters import *
from .utilities import *
from .windows import *
from .sync import *
from .metrics import *
from .multirate import *
from .acoustics import *
from .recognition import *

from . import doa
from . import adaptive
from . import transform
from . import transform as realtime  # to be deprecated
from . import experimental
from . import datasets
from . import bss
from . import denoise
from . import phase

import warnings
warnings.warn(
    "Do not import the 'realtime' module as it will be deprecated soon. It will be replaced by the 'transform' module."
    , DeprecationWarning
)



# @version: 1.0  date: 2017/06/20 by Robin Scheibler
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2017

from unittest import TestCase
import numpy as np
from scipy.signal import fftconvolve

import pyroomacoustics as pra

# fix the RNG seed for repeatability
np.random.seed(0)

# Location of original source
azimuth = 61. / 180. * np.pi  # 60 degrees
tol = 0.3 / 180. * np.pi  # 0.3 degrees tolerance for the test

# algorithms parameters
c = 343.
fs = 16000
nfft = 256
freq_bins = np.arange(5, 60)

# circular microphone array, 12 mics, radius 15 cm
R = pra.circular_2D_array([0, 0], 12, 0., 0.15)

# propagation filter bank
propagation_vector = -np.array([np.cos(azimuth), np.sin(azimuth)])
delays = np.dot(R.T, propagation_vector) / c * fs  # in fractional samples
filter_bank = pra.fractional_delay_filter_bank(delays)

# we use a white noise signal for the source
x = np.random.randn((nfft // 2 + 1) * nfft)

# convolve the source signal with the fractional delay filters
# to get the microphone input signals
mic_signals = np.array([ fftconvolve(x, filter, mode='same') for filter in filter_bank ])
X = pra.transform.analysis(mic_signals.T, nfft, nfft // 2, win=np.hanning(nfft))
X = np.swapaxes(X, 2, 0)

class TestDOA(TestCase):

    def test_music(self):
        doa = pra.doa.algorithms['MUSIC'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_srp_phat(self):
        doa = pra.doa.algorithms['SRP'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_cssm(self):
        doa = pra.doa.algorithms['CSSM'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_tops(self):
        doa = pra.doa.algorithms['TOPS'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_waves(self):
        doa = pra.doa.algorithms['WAVES'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

    def test_frida(self):
        doa = pra.doa.algorithms['FRIDA'](R, fs, nfft, c=c)
        doa.locate_sources(X, freq_bins=freq_bins)
        print('distance:', pra.doa.circ_dist(azimuth, doa.azimuth_recon))
        self.assertTrue(pra.doa.circ_dist(azimuth, doa.azimuth_recon) < tol)

if __name__ == '__main__':

    algo_names = sorted(pra.doa.algorithms.keys())

    for algo_name in algo_names:
        doa = pra.doa.algorithms[algo_name](R, fs, nfft, c=c, max_four=4)
        doa.locate_sources(X, freq_bins=freq_bins)
        print(algo_name, doa.azimuth_recon / np.pi * 180., 
            pra.doa.circ_dist(azimuth, doa.azimuth_recon) / np.pi * 180.)
